def get_partial_id_fraction(reference, target, start, end):
    matches = 0
    # print("reference[start:end]: ", reference[start:end])
    # print("target[start:end]   : ", target[start:end])
    reference = reference.upper()
    target = target.upper()
    gaps_in_ref = 0
    for i, letter in enumerate(reference[start:end]):
        if letter == '-':
            gaps_in_ref += 1
        if letter == target[i+start]:
            matches += 1
        if target[i+start] == "*":
            break
    # Modify the region length by considering gaps in the reference (as long as it's a open reading frame)
    total_length = (end-start) - gaps_in_ref
    if total_length == 0:
        return matches, 1
    return matches, total_length


# Gap-collapsed protein sequence identity
def get_AA_id_fraction(reference, target):
    matches = 0
    # gap-compressed BLAST identity
    reference = reference.upper()
    target = target.upper()
    gaps_in_ref = 0
    for i, letter in enumerate(reference):
        if letter == '-':
            gaps_in_ref += 1
        if letter == target[i]:
            matches += 1
        if target[i] == "*":
            break        
    if max(len(reference), len(target)) == 0:  
        return matches, 1
    # Modify the region length by considering gaps in the reference (as long as it's a open reading frame)
    total_length = max(len(reference), len(target)) - gaps_in_ref    
    return matches, total_length


def get_DNA_id_fraction(reference, target):
    matches = 0
    # BLAST identity
    reference = reference.upper()
    target = target.upper()
    for i, letter in enumerate(reference):
        if letter == target[i]:
            matches += 1
    if max(len(reference), len(target)) == 0:  
        return matches, 1
    return matches, max(len(reference), len(target))


import re, sys, os, copy
from Bio.Seq import Seq
from lifton import align, lifton_class, run_liftoff, run_miniprot, logger
from lifton.liftoff import liftoff_main

def check_miniprot_installed():
    """
        This function checks if miniprot is installed.

        Parameters:
        None

        Returns:
        Exist the program if miniprot is not installed.
    """
    miniprot_installed = run_miniprot.check_miniprot_installed()
    # print("miniprot_installed: ", miniprot_installed)
    if not miniprot_installed:
        if not miniprot_installed:
            print("miniprot is not properly installed.")
        return sys.exit(1)


def get_truncated_protein(ref_proteins):
    """
        This function gets the truncated proteins.

        Parameters:
        - ref_proteins: reference proteins dictionary

        Returns:
        truncated_proteins: truncated proteins dictionary
    """
    truncated_proteins = {}
    for record in ref_proteins.keys():
        protein = ref_proteins[record]
        if not check_protein_valid(str(protein)):
            truncated_proteins[record] = protein
    # print("truncated_proteins: ", len(truncated_proteins))
    # print("good_protein: ", good_protein)
    # print("bad_protein: ", bad_protein)
    return truncated_proteins


def write_seq_2_file(outdir, ref_seqs, target):
    """
        This function writes the reference sequences to a file.

        Parameters:
        - outdir: output directory
        - ref_seqs: reference sequences dictionary
        - target: target type ('truncated_proteins', 'proteins', or 'transcripts')

        Returns:
        ref_seqs_file: reference sequences file path
    """
    if target == "truncated_proteins":
        ref_seqs_file = outdir + "/proteins_truncated.fa"
    elif target == "proteins":
        ref_seqs_file = outdir + "/proteins.fa"
    elif target == "transcripts":
        ref_seqs_file = outdir + "/transcripts.fa"
    fw = open(ref_seqs_file, 'w')
    # Iterate through the original FASTA and write the records to the new FASTA file
    for record in ref_seqs.keys():
        seq = ref_seqs[record]
        fw.write(f'>{record}\n{seq}\n')
    fw.close()
    return ref_seqs_file


def check_protein_valid(protein):
    """
        This function checks if the protein is valid.

        Parameters:
        - protein: protein sequence

        Returns:
        True if the protein is valid, False otherwise.
    """
    # The length of the protein has to be greater than 0
    if len(protein) == 0:
        return False
    # Start with M
    if protein[0] != "M":
        return False
    # End with *
    if protein[-1] != "*":
        return False
    # Only 1 * in the string
    if protein.count("*") != 1:
        return False
    return True
    

def exec_liftoff(outdir, ref_db, args):
    """
        This function executes liftoff.

        Parameters:
        - outdir: output directory
        - args: arguments

        Returns:
        liftoff_annotation: liftoff annotation file path
    """
    liftoff_annotation = args.liftoff
    if liftoff_annotation is None or not os.path.exists(liftoff_annotation):
        print("\n*********************")
        print("** Running Liftoff **")
        print("*********************")
        liftoff_annotation = run_liftoff.run_liftoff(outdir, ref_db, args)
    return liftoff_annotation


def exec_miniprot(outdir, args, tgt_genome, ref_proteins_file):
    """
        This function executes miniprot.

        Parameters:
        - outdir: output directory
        - args: arguments
        - tgt_genome: target genome
        - ref_proteins_file: reference protein file

        Returns:
        miniprot_annotation: miniprot annotation file path
    """
    check_miniprot_installed()
    miniprot_annotation = args.miniprot
    if miniprot_annotation is None or not os.path.exists(miniprot_annotation):
        print("\n**********************")
        print("** Running miniprot **")
        print("**********************")
        miniprot_annotation = run_miniprot.run_miniprot(outdir, args, tgt_genome, ref_proteins_file)
    return miniprot_annotation


def custom_bisect_insert(sorted_list, element_to_insert):
    """
        This function bisects the sorted list and inserts the element.

        Parameters:
        - sorted_list: sorted list
        - element_to_insert: element to insert

        Returns:
        None
    """
    low = 0
    high = len(sorted_list)

    while low < high:
        mid = (low + high) // 2
        if sorted_list[mid].entry.end < element_to_insert.entry.end:
            low = mid + 1
        else:
            high = mid
    sorted_list.insert(low, element_to_insert)


def get_ID_base(id, ref_features_dict=None):
    """
        This function gets the ID base by removing the last substring after "_" if it's a copy number.
        Only removes the suffix if it's confirmed to be a copy number added by LiftOn.
        
        The key insight: We can only safely remove a suffix if the ID without the suffix exists
        in the reference features dictionary. This confirms the suffix was added by LiftOn,
        not part of the original ID.

        Parameters:
        - id: ID
        - ref_features_dict: Optional reference features dictionary to verify if the base ID exists.
                            If provided, only removes suffix if base ID exists in the dictionary.

        Returns:    
        ID base (original ID if suffix removal is not safe or cannot be verified)
    """
    splits = id.split("_")
    # Only try to remove suffix if there are at least 2 parts after splitting
    if len(splits) < 2:
        return id
    
    # Check if the last part is a pure integer (potential copy number)
    try:
        last_part = splits[-1]
        copy_num = int(last_part)
        id_base = "_".join(splits[:-1])
        
        # If ref_features_dict is provided, verify that removing the suffix gives a valid ID
        # This is the safest approach: only remove if we can confirm the base exists
        if ref_features_dict is not None:
            # Only remove the suffix if the base ID exists in the reference dictionary
            # This confirms the suffix was added by LiftOn, not part of the original ID
            if id_base in ref_features_dict.keys():
                return id_base
            else:
                # The suffix is likely part of the original ID (e.g., FMUND_1), don't remove it
                return id
        else:
            # Without reference dict, we cannot safely determine if the suffix is a copy number
            # Be conservative: don't remove the suffix to avoid breaking IDs that naturally end with numbers
            # This prevents issues like GCA_013396205.1-transcript_rna-gnl-WGS:JAAOAN-mrna.FMUND_1
            # where _1 is part of the original ID, not a copy number
            return id
    except (ValueError, IndexError):
        # Last part is not an integer, so it's not a copy number suffix
        return id


def get_ID(feature):
    """
        This function gets the ID.

        Parameters:
        - feature: feature

        Returns:    
        the original ID and the ID base
    """
    # print("feature: ", feature)
    # id_spec={"ID"}
    id = feature.id
    id_base = get_ID_base(id)
    return id, id_base


def get_parent_features_to_lift(feature_types_file):
    """
        This function gets the parent features to lift.

        Parameters:
        - feature_types_file: feature types file (from '-f' / '--features')

        Returns:
        feature_types: feature types
    """
    feature_types = ["gene"]
    if feature_types_file is not None:
        feature_types = []
        f = open(feature_types_file)
        for line in f.readlines():
            feature_types.append(line.rstrip())
    return feature_types


def LiftOn_eval_alignment(eval_trans, locus, tgt_fai, ref_proteins, ref_trans_id, lifton_status):
    eval_aln = align.lifton_parasail_align(eval_trans, locus, tgt_fai, ref_proteins, ref_trans_id)
    if eval_aln != None:
        lifton_status.lifton_aa = eval_aln.identity
    return eval_aln


def LiftOn_liftoff_alignment(lifton_trans, locus, tgt_fai, ref_proteins, ref_trans_id, lifton_status):
    liftoff_aln = align.lifton_parasail_align(lifton_trans, locus, tgt_fai, ref_proteins, ref_trans_id)
    if liftoff_aln != None:
        lifton_status.liftoff = liftoff_aln.identity
    return liftoff_aln


def LiftOn_miniprot_alignment(chromosome, transcript, m_id_dict, m_feature_db, tree_dict, fai, ref_proteins, ref_trans_id, lifton_status):
    """
        This function checks the miniprot alignment.

        Parameters:
        - chromosome: chromosome
        - transcript: transcript gffutils feature
        - lifton_status: Lifton_Status instance
        - m_id_dict: miniprot id dictionary
        - m_feature_db: miniprot feature database
        - tree_dict: tree dictionary
        - fai: reference fasta index
        - ref_proteins: reference proteins dictionary
        - ref_trans_id: reference transcript ID

        Returns:
        m_lifton_aln: miniprot lifton alignment
        has_valid_miniprot: True if the miniprot transcript is valid, False otherwise.
    """
    m_lifton_aln = None
    has_valid_miniprot = False
    if (ref_trans_id in m_id_dict.keys()) and (ref_trans_id in ref_proteins.keys()):
        m_ids = m_id_dict[ref_trans_id]
        for m_id in m_ids:
            ##################################################
            # Check 1: Check if the miniprot transcript is overlapping the current gene locus
            ##################################################
            m_entry = m_feature_db[m_id]
            _, overlap = segments_overlap_length((m_entry.start, m_entry.end), (transcript.start, transcript.end))
            if not overlap or m_entry.seqid != transcript.seqid:
                # "Not overlapped"
                continue
            ##################################################
            # Check 2: reference overlapping status
            #   1. Check it the transcript overlapping with the next gene
            # Check the miniprot protein overlapping status
            # The case I should not process the transcript 
            #   1. The Liftoff does not overlap with other gene
            #   2. The miniprot protein overlap the other gene
            ##################################################
            ovps_liftoff = tree_dict[chromosome].overlap(transcript.start, transcript.end)
            ovps_miniprot = tree_dict[chromosome].overlap(m_entry.start, m_entry.end)
            miniprot_cross_gene_loci = False
            liftoff_set = set()
            for ovp_liftoff in ovps_liftoff:
                liftoff_set.add(ovp_liftoff[2])
            for ovp_miniprot in ovps_miniprot:
                if ovp_miniprot[2] not in liftoff_set:
                    # Miniprot overlap to more genes
                    miniprot_cross_gene_loci = True
                    break
            if miniprot_cross_gene_loci:
                continue
            # Valid miniprot transcript exists => check if the miniprot transcript is valid
            has_valid_miniprot = True
            miniprot_trans = lifton_class.Lifton_TRANS(m_id, "", "", 0, m_entry, {})
            exons = m_feature_db.children(m_entry, featuretype=('CDS', 'stop_codon'), order_by='start')
            for exon in list(exons):
                miniprot_trans.add_exon(exon)
            cdss = m_feature_db.children(m_entry, featuretype=('CDS', 'stop_codon'), order_by='start') 
            cds_num = 0
            for cds in list(cdss):
                cds_num += 1
                miniprot_trans.add_cds(cds)
            tmp_m_lifton_aln = align.lifton_parasail_align(miniprot_trans, m_entry, fai, ref_proteins, ref_trans_id)
            if m_lifton_aln == None or tmp_m_lifton_aln.identity > lifton_status.miniprot:
                m_lifton_aln = tmp_m_lifton_aln
                lifton_status.miniprot = m_lifton_aln.identity
    return m_lifton_aln, has_valid_miniprot


def get_ref_liffover_features(features, ref_db, intermediate_dir, args):
    """
        This function gets the reference liftover features.

        Parameters:
        - features: list of features to liftover
        - ref_db: reference database

        Returns:
        ref_features_dict: reference features dictionary (gene id -> transcript id)
        ref_features_reverse_dict: reference features reverse dictionary (transcript id -> gene id)
    """

    fw_gene = open(f"{intermediate_dir}/ref_feature.txt", "w")
    fw_trans = open(f'{intermediate_dir}/ref_transcript.txt', 'w')    
    ref_features_dict = {}
    ref_features_len_dict = {}
    ref_features_reverse_dict = {}
    ref_trans_exon_num_dict = {}
    new_gene_feature = lifton_class.Lifton_feature("Lifton-gene")
    ref_features_dict["LiftOn-gene"] = new_gene_feature
    for f_itr in features:
        for locus in ref_db.db_connection.features_of_type(f_itr):
            CDS_children = list(ref_db.db_connection.children(locus, featuretype='CDS'))
            feature = lifton_class.Lifton_feature(locus.id)
            # Write out reference gene features IDs
            # Decide if its type
            # Check for gene type/biotype attribute in order of preference
            gene_type_key = None
            if args.annotation_database.upper() == "REFSEQ":
                # For RefSeq, check gene_biotype first, then biotype as fallback
                if "gene_biotype" in locus.attributes.keys():
                    gene_type_key = "gene_biotype"
                elif "biotype" in locus.attributes.keys():
                    gene_type_key = "biotype"
            elif args.annotation_database.upper() == "GENCODE" or args.annotation_database.upper() == "ENSEMBL" or args.annotation_database.upper() == "CHESS":
                # For GENCODE/ENSEMBL/CHESS, check gene_type first, then biotype as fallback
                if "gene_type" in locus.attributes.keys():
                    gene_type_key = "gene_type"
                elif "biotype" in locus.attributes.keys():
                    gene_type_key = "biotype"
            else:
                # For other databases, try all possible keys in order
                if "gene_biotype" in locus.attributes.keys():
                    gene_type_key = "gene_biotype"
                elif "gene_type" in locus.attributes.keys():
                    gene_type_key = "gene_type"
                elif "biotype" in locus.attributes.keys():
                    gene_type_key = "biotype"

            if gene_type_key is not None:
                if locus.attributes[gene_type_key][0] == "protein_coding" and len(CDS_children) > 0:
                    feature.is_protein_coding = True
                    fw_gene.write(f"{locus.id}\tcoding\n")
                elif (locus.attributes[gene_type_key][0] == "lncRNA" or locus.attributes[gene_type_key][0] == "ncRNA"):
                    feature.is_non_coding = True
                    fw_gene.write(f"{locus.id}\tnon-coding\n")
                else:
                    fw_gene.write(f"{locus.id}\tother\n")
            else:
                fw_gene.write(f"{locus.id}\tother\n")
            exon_children = list(ref_db.db_connection.children(locus, featuretype='exon', level=1, order_by='start'))
            if len(exon_children) > 0:
                __process_ref_liffover_features(locus, ref_db, None)
            else:
                transcripts = ref_db.db_connection.children(locus, level=1)
                for transcript in list(transcripts):
                    __process_ref_liffover_features(transcript, ref_db, feature)
                    ref_features_reverse_dict[transcript.id if not args.evaluation_liftoff_chm13 else locus.id[4:]] = locus.id
                    all_CDS_in_trans = list(ref_db.db_connection.children(transcript, featuretype='CDS', order_by='start'))
                    if len(all_CDS_in_trans) > 0:
                        ref_trans_exon_num_dict[transcript.id if not args.evaluation_liftoff_chm13 else locus.id[4:]] = len(all_CDS_in_trans)
                    else:
                        ref_trans_exon_num_dict[transcript.id if not args.evaluation_liftoff_chm13 else locus.id[4:]] = 0
                    # Write out reference trans feature IDs
                    if feature.is_protein_coding and transcript.featuretype == "mRNA":
                        fw_trans.write(f"{transcript.id}\tcoding\n")
                    elif feature.is_non_coding and (transcript.featuretype == "ncRNA" or transcript.featuretype == "nc_RNA" or transcript.featuretype == "lncRNA" or transcript.featuretype == "lnc_RNA"):
                        fw_trans.write(f"{transcript.id}\tnon-coding\n")
                    else:
                        fw_trans.write(f"{transcript.id}\tother\n")
            ref_features_dict[locus.id if not args.evaluation_liftoff_chm13 else locus.id[5:]] = feature
            all_CDS_children = list(ref_db.db_connection.children(locus, featuretype='CDS', order_by='start'))
            if len(all_CDS_children) > 0:
                ref_features_len_dict[locus.id if not args.evaluation_liftoff_chm13 else locus.id[5:]] = all_CDS_children[-1].end - all_CDS_children[0].start + 1
            else:
                ref_features_len_dict[locus.id if not args.evaluation_liftoff_chm13 else locus.id[5:]] = 0
    fw_gene.close()
    fw_trans.close()
    return ref_features_dict, ref_features_len_dict, ref_features_reverse_dict, ref_trans_exon_num_dict


def __process_ref_liffover_features(locus, ref_db, feature):
    if feature != None:
        feature.children.add(locus.id)


def miniprot_id_mapping(m_feature_db):
    """
        This function creates a dictionary of miniprot id to reference id.

        Parameters:
        - m_feature_db: miniprot feature database

        Returns:
        ref_id_2_m_id_trans_dict: reference id to miniprot transcript ids dictionary
        m_id_2_ref_id_trans_dict: miniprot transcript id to reference id dictionary
    """
    ref_id_2_m_id_trans_dict = {}
    m_id_2_ref_id_trans_dict = {}
    for feature in m_feature_db.features_of_type("mRNA"):
        miniprot_id = feature["ID"][0]
        aa_trans_id = str(feature.attributes["Target"][0]).split(" ")[0]
        if aa_trans_id in ref_id_2_m_id_trans_dict.keys():
            ref_id_2_m_id_trans_dict[aa_trans_id].append(miniprot_id)
        else:
            ref_id_2_m_id_trans_dict[aa_trans_id] = [miniprot_id]
        m_id_2_ref_id_trans_dict[miniprot_id] = aa_trans_id
    return ref_id_2_m_id_trans_dict, m_id_2_ref_id_trans_dict


def get_ref_ids_liftoff(ref_features_dict, liftoff_gene_id, liftoff_trans_id):
    """
        This function gets the reference IDs from Liftoff IDs.

        Parameters:
        - ref_features_dict: reference features dictionary
        - liftoff_gene_id: Liftoff gene ID
        - liftoff_trans_id: Liftoff transcript ID

        Returns:
        # Three cases that will be passed into this function
        1.  Liftoff_gene_id, None
        2.  None, Liftoff_trans_id
        3.  Liftoff_gene_id, Liftoff_trans_id
    """
    if liftoff_gene_id is None:
        ref_trans_id = __extract_ref_ids(ref_features_dict, liftoff_trans_id)
        return None, ref_trans_id 

    if liftoff_trans_id is None:
        ref_gene_id = __extract_ref_ids(ref_features_dict, liftoff_gene_id)
        return ref_gene_id, None

    ref_gene_id = __extract_ref_ids(ref_features_dict, liftoff_gene_id)
    if ref_gene_id is None:
        return None, None
    else:
        # For transcript ID, we need to check if the base exists in ref_features_dict
        # But ref_features_dict contains gene IDs, not transcript IDs
        # So we'll use a more conservative approach: only remove single-digit suffixes
        # The transcript ID should match the reference transcript ID from the reference annotation
        ref_trans_id = get_ID_base(liftoff_trans_id, None)
        return ref_gene_id, ref_trans_id


def __extract_ref_ids(ref_features_dict, liftoff_id):
    if liftoff_id in ref_features_dict.keys():
        return liftoff_id
    else:
        # Pass ref_features_dict to get_ID_base so it can verify the base ID exists
        ref_id = get_ID_base(liftoff_id, ref_features_dict)
        if ref_id in ref_features_dict.keys():
            return ref_id
        else:
            return None


def get_ref_ids_miniprot(ref_features_reverse_dict, miniprot_trans_id, m_id_2_ref_id_trans_dict):
    if miniprot_trans_id not in m_id_2_ref_id_trans_dict.keys():
        return None, None
    ref_trans_id = m_id_2_ref_id_trans_dict[miniprot_trans_id]
    if ref_trans_id not in ref_features_reverse_dict.keys():
        return None, ref_trans_id
    return ref_features_reverse_dict[ref_trans_id], ref_trans_id


def write_lifton_eval_status(fw_score, transcript_id, transcript, lifton_status):
    final_status = ";".join(lifton_status.status)
    fw_score.write(f"{transcript_id}\t{lifton_status.eval_dna}\t{lifton_status.eval_aa}\t{lifton_status.annotation}\t{final_status}\t{transcript.seqid}:{transcript.start}-{transcript.end}\n")


def print_lifton_status(transcript_id, transcript, lifton_status, DEBUG=False):
    final_status = ";".join(lifton_status.status)
    logger.log(f"{transcript_id}\t{lifton_status.liftoff}\t{lifton_status.miniprot}\t{lifton_status.lifton_dna}\t{lifton_status.lifton_aa}\t{lifton_status.annotation}\t{final_status}\t{transcript.seqid}:{transcript.start}-{transcript.end}\n", debug=DEBUG)


def write_lifton_status(fw_score, transcript_id, transcript, lifton_status):
    final_status = ";".join(lifton_status.status)
    fw_score.write(f"{transcript_id}\t{lifton_status.liftoff}\t{lifton_status.miniprot}\t{lifton_status.lifton_dna}\t{lifton_status.lifton_aa}\t{lifton_status.annotation}\t{final_status}\t{transcript.seqid}:{transcript.start}-{transcript.end}\n")


def write_lifton_eval_status(fw_score, transcript_id, transcript, lifton_status):
    final_status = ";".join(lifton_status.status)
    fw_score.write(f"{transcript_id}\t{lifton_status.lifton_dna}\t{lifton_status.lifton_aa}\t{final_status}\t{transcript.seqid}:{transcript.start}-{transcript.end}\n")


def write_lifton_chains(fw_chain, transcript_id, chains):
    chain_ls = ";".join(chains)
    fw_chain.write(f"{transcript_id}\t{chain_ls}\n")


def segments_overlap_length(segment1, segment2):
    """
        This function gets the length of the overlapping segments.

        Parameters:
        - segment1: segment 1 in tuple (start, end)
        - segment2: segment 2 in tuple (start, end)

        Returns:
        The length of the overlapping segments.
    """
    if len(segment1) != 2 or len(segment2) != 2:
        raise ValueError("Segments must have exactly 2 endpoints")
    # Sort the segments by their left endpoints
    segment1, segment2 = sorted([segment1, segment2], key=lambda x: x[0])
    ovp_len = segment1[1] - segment2[0] + 1
    ovp = False
    if ovp_len > 0: ovp = True
    return ovp_len, ovp


def check_ovps_ratio(mtrans, mtrans_interval, overlap_ratio, tree_dict):
    """
        This function checks the overlap ratio.

        Parameters:
        - mtrans: miniprot transcript gffutils feature
        - mtrans_interval: miniprot transcript interval
        - overlap_ratio: overlap ratio
        - tree_dict: tree dictionary

        Returns:
        True if the overlap ratio is greater than the threshold, False otherwise.
    """
    is_overlapped = False
    if mtrans.seqid not in tree_dict.keys():
        return False
    ovps = tree_dict[mtrans.seqid].overlap(mtrans_interval)
    for ovp in ovps:
        ovp_len, _ = segments_overlap_length((mtrans_interval[0], mtrans_interval[1]), (ovp[0], ovp[1]))
        ref_len = ovp[1] - ovp[0] + 1
        target_len = mtrans_interval[1] - mtrans_interval[0] + 1
        # Overlapping does not extend the ratio of the reference
        if (ovp_len / min(ref_len, target_len)) > overlap_ratio:
            is_overlapped = True
            break
    return is_overlapped


from glob import glob
import lifton.liftoff 
from lifton.liftoff import liftoff_main
import os
import pytest

def test_yeast(tmp_path):
    os.makedirs("./"+tmp_path, exist_ok=True)
    # cleanup
    for tfile in glob('lifton/liftoff/tests/*.mmi') + glob('lifton/liftoff/tests/*.fai') + glob('lifton/liftoff/tests/*_db'):
        print("tfile: ", tfile)
        try:
            os.unlink(tfile)
        except OSError:
            pass

    # inputs
    asmbl = 'lifton/liftoff/tests/GCA_000146045.2_R64_genomic.fna.gz'
    annot = 'lifton/liftoff/tests/GCA_000146045.2_R64_genomic.gff.gz'
    target = 'lifton/liftoff/tests/GCA_000146045.2_R64_genomic.fna.gz'
    chroms = 'lifton/liftoff/tests/chroms.txt'
    unplaced = 'lifton/liftoff/tests/unplaced.txt'

    print("tmp_path: ", tmp_path)
    # outputs
    output = f'{tmp_path}/GCA_000146045.2_R64_to_GCA_000146045.2_advanced.gff3'
    unmapped = f'{tmp_path}/unmapped_features_GCA_000146045.2_R64_advanced.txt'
    tempdir = f'{tmp_path}/sandbox'
    expout = 'lifton/liftoff/tests/GCA_000146045.2_R64_to_GCA_000146045.2_R64_expected_advanced.gff'
    expunmapped = 'lifton/liftoff/tests/unmapped_features_GCA_000146045.2_R64_expected_advanced.txt'

    # run the program
    args = ['-g', annot, '-o', output, '-u', unmapped, '-dir', tempdir,
            '-chroms', chroms, '-unplaced', unplaced, '-copies',  target, asmbl]
    liftoff_main.main(args)

    # verify the output
    with open(output, 'r') as fh1, open(expout, 'r') as fh2:
        fh1_lines = fh1.readlines()
        fh2_lines = fh2.readlines()
        for i in range(len(fh1_lines)):
            if fh1_lines[0][0] != "#":
                observed_output = fh1_lines[i].strip().split("\t")[0:8]
                expected_output = fh2_lines[i].strip().split("\t")[0:8]
                assert observed_output == expected_output
    with open(unmapped, 'r') as fh1, open(expunmapped, 'r') as fh2:
        observed_output = fh1.read().strip()
        expected_output = fh2.read().strip()
        assert observed_output == expected_output



import networkx as nx
from lifton.liftoff  import aligned_seg, liftoff_utils, new_feature
import numpy as np


def find_best_mapping(alignments, query_length, parent, feature_heirarchy, previous_feature_start,
                      previous_feature_ref_start, previous_gene_seq,
                      inter,lifted_features_list, args):
    children = feature_heirarchy.children[parent.id]
    children_coords = liftoff_utils.merge_children_intervals(children)
    node_dict, aln_graph = intialize_graph()
    head_nodes = add_single_alignments(node_dict, aln_graph, alignments, children_coords, parent,
                                       previous_feature_start, previous_feature_ref_start, previous_gene_seq, inter,
                                       feature_heirarchy.parents,
                                       lifted_features_list, args)

    chain_alignments(head_nodes, node_dict, aln_graph, parent, children_coords, inter,
                     feature_heirarchy.parents, lifted_features_list, args,
                     )
    add_target_node(aln_graph, node_dict, query_length, children_coords, parent, args)
    shortest_path_nodes = find_shortest_path(node_dict, aln_graph)
    if len(shortest_path_nodes) == 0:
        return {}, 0, 0
    mapped_children, alignment_coverage, seq_id = convert_all_children_coords(shortest_path_nodes, children, parent)
    return mapped_children, alignment_coverage, seq_id


def intialize_graph():
    aln_graph = nx.DiGraph()
    node_dict = {}
    node_dict[0] = aligned_seg.aligned_seg("start", "start", "start", -1, -1, -1, -1, -1, [])
    aln_graph.add_node(0)
    return node_dict, aln_graph


def add_single_alignments(node_dict, aln_graph, alignments, children_coords, parent,
                          previous_feature_start, previous_feature_ref_start, previous_gene_seq, inter, parent_dict,
                          lifted_features_list, args
                          ):
    if inter is not None:
        alignments = remove_alignments_with_overlap(alignments, inter, parent, parent_dict, lifted_features_list)
    alignments.sort(key=lambda x: x.query_block_start)
    alignments = sort_alignments(parent, previous_feature_start, previous_feature_ref_start, previous_gene_seq,
                                 alignments)

    node_num = 1
    previous_node_id = -1
    head_nodes = []
    previous_node = 0
    for aln in alignments:
        if aln.aln_id != previous_node_id:
            previous_node = 0
            previous_node_id = aln.aln_id
        if is_valid_alignment(previous_node, aln, node_dict, parent, inter, parent_dict,
                              lifted_features_list):
            if previous_node == 0:
                head_nodes.append(node_num)
            current_node = add_to_graph(aln_graph, node_num, children_coords, parent, node_dict, previous_node, aln,
                                        args)
            previous_node = current_node
            node_num += 1
    return head_nodes


def sort_alignments(parent, previous_feature_start, previous_feature_ref_start, previous_gene_seq,
                    alignments):
    order_dict = {}
    order = 0
    alignments = sorted(alignments, key=lambda x: [is_different_chrom(previous_gene_seq, x.reference_name),
                                                   distance_difference(parent.start, x.reference_block_start,
                                                                       previous_feature_ref_start,
                                                                       previous_feature_start, x.query_block_start, )])
    for aln in alignments:
        if aln.aln_id not in order_dict:
            order_dict[aln.aln_id] = order
            order += 1
    alignments = sorted(alignments, key=lambda x: (order_dict[x.aln_id], x.query_block_start))
    return alignments


def remove_alignments_with_overlap(alignments, inter, parent, parent_dict, lifted_features_list):
    aln_ids = set([aln.aln_id for aln in alignments])
    alignments_to_keep = []
    for aln_id in aln_ids:
        single_alignment_group = [aln for aln in alignments if aln.aln_id == aln_id]
        min_ref_start = min([aln.reference_block_start for aln in single_alignment_group])
        max_ref_end = max([aln.reference_block_end for aln in single_alignment_group])
        chrom = single_alignment_group[0].reference_name
        strand = get_strand(single_alignment_group[0], parent)
        feature_name = single_alignment_group[0].query_name
        overlaps = liftoff_utils.find_overlaps(min_ref_start, max_ref_end, chrom, strand, feature_name, inter,
                                               parent_dict, lifted_features_list, 0)
        if len(overlaps) == 0:
            alignments_to_keep += single_alignment_group
    return alignments_to_keep


def is_different_chrom(previous_gene_seq, current_seq):
    return previous_gene_seq != current_seq


def distance_difference(feature_ref_start, feature_target_start, previous_feature_ref_start,
                        previous_feature_target_start, query_start):
    expected_distance = feature_ref_start - previous_feature_ref_start
    actual_distance = feature_target_start - query_start - previous_feature_target_start
    return np.abs(expected_distance - actual_distance)


def get_strand(aln, parent):
    if aln.is_reverse:
        if parent.strand == "-":
            strand = '+'
        else:
            strand = "-"
    else:
        strand = parent.strand
    return strand


def is_valid_alignment(previous_node, aln, node_dict, parent, inter, parent_dict,
                       lifted_features_list):
    if aln.reference_block_start == -1:
        return False
    elif previous_node != 0:
        if spans_overlap_region(node_dict[previous_node], aln, parent, inter, parent_dict, lifted_features_list):
            return False
    return True


def spans_overlap_region(from_node, to_node, parent, intervals, parent_dict, lifted_features_list):
    if from_node.reference_name != to_node.reference_name:
        return False
    if intervals == None:
        return False
    target_chrm = from_node.reference_name
    node_overlap = get_node_overlap(from_node, to_node)
    strand = get_strand(from_node, parent)
    overlaps = liftoff_utils.find_overlaps(from_node.reference_block_end, to_node.reference_block_start + node_overlap,
                                           target_chrm, strand, from_node.query_name, intervals,
                                           parent_dict, lifted_features_list, 0)
    return len(overlaps) > 0


def get_node_overlap(from_node, to_node):
    if from_node.reference_name == "start" or to_node.reference_name == "start":
        return 0
    else:
        return max(0, from_node.query_block_end - to_node.query_block_start + 1)


def add_to_graph(aln_graph, node_num, children_coords, parent, node_dict, previous_node, aln, args):
    node_dict[node_num] = aln
    current_node = node_num
    node_weight = get_node_weight(aln, children_coords, parent, args)
    aln_graph.add_node(node_num, weight=node_weight)
    edge_weight = get_edge_weight(node_dict[previous_node], node_dict[current_node], children_coords, parent, args)
    aln_graph.add_edge(previous_node, current_node, cost=edge_weight)
    return current_node


def get_node_weight(aln, children_coords, parent, args):
    weight = 0
    for child_interval in children_coords:
        relative_start = liftoff_utils.get_relative_child_coord(parent, child_interval[0], aln.is_reverse)
        relative_end = liftoff_utils.get_relative_child_coord(parent, child_interval[1], aln.is_reverse)
        child_start, child_end = min(relative_start, relative_end), max(relative_start, relative_end)
        weight += len(aln.mismatches[(aln.mismatches >= child_start) & (aln.mismatches <= child_end)]) * args.mismatch
    return weight


def get_edge_weight(from_node, to_node, children_coords, parent, args):
    node_overlap = get_node_overlap(from_node, to_node)
    unaligned_range = [from_node.query_block_end + 1, to_node.query_block_start + node_overlap - 1]
    unaligned_exon_bases = 0
    for child_interval in children_coords:
        if from_node.reference_name == "start":
            is_reverse = to_node.is_reverse
        else:
            is_reverse = from_node.is_reverse
        relative_start = liftoff_utils.get_relative_child_coord(parent, child_interval[0], is_reverse)
        relative_end = liftoff_utils.get_relative_child_coord(parent, child_interval[1], is_reverse)
        child_start, child_end = min(relative_start, relative_end), max(relative_start, relative_end)
        overlap = liftoff_utils.count_overlap(child_start, child_end, min(unaligned_range[0], unaligned_range[1]),
                                              max(unaligned_range[0], unaligned_range[1]))
        if overlap == 1 and unaligned_range[0] == unaligned_range[1] + 1 and from_node.reference_name == \
                to_node.reference_name:
            unaligned_exon_bases += ((to_node.reference_block_start + node_overlap) - from_node.reference_block_end \
                                     - 1) * args.gap_extend
        else:
            unaligned_exon_bases += max(0, overlap) * args.gap_extend
    if unaligned_exon_bases > 0:
        unaligned_exon_bases += (args.gap_open - args.gap_extend)  # gap open penalty
    return unaligned_exon_bases


def chain_alignments(head_nodes, node_dict, aln_graph, parent, children_coords, intervals,
                     parent_dict, lifted_features_list, args):
    for node_name in head_nodes:
        add_edges(node_name, node_dict, aln_graph, parent, children_coords, intervals,
                  parent_dict, lifted_features_list, args)


def add_edges(head_node_name, node_dict, aln_graph, parent, children_coords, intervals,
              parent_dict, lifted_features_list, args):
    for node_name in node_dict:
        from_node = node_dict[node_name]
        if is_valid_edge(node_name, head_node_name, parent, intervals,
                         parent_dict, lifted_features_list, args
                , node_dict):
            edge_weight = get_edge_weight(from_node, node_dict[head_node_name], children_coords, parent, args)
            aln_graph.add_edge(node_name, head_node_name, cost=edge_weight)


def is_valid_edge(from_node_name, to_node_name, parent, intervals, parent_dict, lifted_features_list, args,
                  node_dict):
    from_node, to_node = node_dict[from_node_name], node_dict[to_node_name]
    if from_node.aln_id == to_node.aln_id:
        return False
    if from_node.query_block_end >= to_node.query_block_end:
        return False
    if from_node.is_reverse != to_node.is_reverse:
        return False
    if from_node.reference_name != to_node.reference_name:
        return False
    else:
        expected_distance = to_node.query_block_end - from_node.query_block_start
        actual_distance = to_node.reference_block_end - from_node.reference_block_start
        if to_node.reference_block_start < from_node.reference_block_end:
            return False
        if (actual_distance > args.d * expected_distance):
            return False
        if spans_overlap_region(from_node, to_node, parent, intervals, parent_dict,
                                lifted_features_list):
            return False
    return True


def add_target_node(aln_graph, node_dict, query_length, children_coords, parent, args):
    num_nodes = max(node_dict.keys())
    node_dict[num_nodes + 1] = aligned_seg.aligned_seg("end", "end", "end", query_length, query_length,
                                                       query_length,
                                                       query_length, 0, [])
    aln_graph.add_node(num_nodes + 1)
    for node_name in node_dict:
        if is_terminal_node(node_name, aln_graph) and node_name != num_nodes + 1:
            edge_weight = get_edge_weight(node_dict[node_name], node_dict[num_nodes + 1], children_coords, parent, args)
            aln_graph.add_edge(node_name, num_nodes + 1, cost=edge_weight)


def is_terminal_node(node, aln_graph):
    successors = aln_graph.successors(node)
    for successor in successors:
        return False
    return True


def find_shortest_path(node_dict, aln_graph):
    shortest_path = nx.shortest_path(aln_graph, source=0, target=len(node_dict) - 1,
                                     weight=lambda u, v, d: get_weight(u, v, d, aln_graph))
    shortest_path_nodes = []
    for i in range(1, len(shortest_path) - 1):
        node_name = shortest_path[i]
        shortest_path_nodes.append(node_dict[node_name])
    trim_path_boundaries(shortest_path_nodes)
    return shortest_path_nodes


def get_weight(u, v, d, G):
    node_u_wt = G.nodes[u].get('weight', 0)
    node_v_wt = G.nodes[v].get('weight', 0)
    edge_wt = d.get('cost', 1)
    return node_u_wt / 2 + node_v_wt / 2 + edge_wt


def trim_path_boundaries(shortest_path_nodes):
    for i in range(1, len(shortest_path_nodes)):
        from_node = shortest_path_nodes[i - 1]
        to_node = shortest_path_nodes[i]
        node_overlap = get_node_overlap(from_node, to_node)
        to_node.query_block_start += node_overlap
        to_node.reference_block_start += node_overlap


def convert_all_children_coords(shortest_path_nodes, children, parent):
    shortest_path_nodes.sort(key=lambda x: x.query_block_start)
    mapped_children = {}
    total_bases, mismatches, insertions, deletions, matches = 0, 0, 0, 0, 0
    for child in children:
        total_bases += (child.end - child.start + 1)
        nearest_start_coord, nearest_end_coord,relative_start, relative_end = find_nearest_aligned_start_and_end(child.start, child.end,shortest_path_nodes, parent)
        if nearest_start_coord != -1 and nearest_end_coord != -1:
            lifted_start, start_node, = convert_coord(nearest_start_coord, shortest_path_nodes)
            lifted_end, end_node = convert_coord(nearest_end_coord, shortest_path_nodes)
            deletions += find_deletions(start_node, end_node, shortest_path_nodes)
            deletions += (nearest_start_coord - relative_start) + (relative_end- nearest_end_coord)
            mismatches += find_mismatched_bases(child.start, child.end, shortest_path_nodes, parent)
            insertions += find_insertions(start_node, end_node, shortest_path_nodes)
            strand = get_strand(shortest_path_nodes[0], parent)
            if "ID" not in child.attributes:
                child.attributes["ID"] = [child.id]
            new_child = new_feature.new_feature(child.id, child.featuretype, shortest_path_nodes[0].reference_name,
                                                'Liftoff',
                                                strand, min(lifted_start, lifted_end) + 1,
                                                max(lifted_start, lifted_end) + 1, child.frame, dict(child.attributes))
            mapped_children[new_child.id] = new_child

        else:
            deletions += (child.end - child.start + 1)
    alignment_length = total_bases + insertions
    return mapped_children, (total_bases - deletions) / total_bases, (alignment_length - insertions - mismatches -
                                                                      deletions) / alignment_length


def find_nearest_aligned_start_and_end(child_start, child_end, shortest_path_nodes, parent):
    relative_coord1 = liftoff_utils.get_relative_child_coord(parent, child_start, shortest_path_nodes[0].is_reverse)
    relative_coord2 = liftoff_utils.get_relative_child_coord(parent, child_end, shortest_path_nodes[0].is_reverse)
    relative_start, relative_end = min(relative_coord1, relative_coord2), max(relative_coord1, relative_coord2)
    nearest_start = find_nearest_aligned_start(relative_start, relative_end, shortest_path_nodes)
    nearest_end = find_nearest_aligned_end(shortest_path_nodes, relative_end, relative_start)
    return nearest_start, nearest_end, relative_start, relative_end


def find_nearest_aligned_start(relative_start, relative_end, shortest_path_nodes):
    nearest_start = -1
    for node in shortest_path_nodes:
        if relative_start <= node.query_block_end:
            if relative_start >= node.query_block_start:
                nearest_start = relative_start
            else:
                if node.query_block_start < relative_end:
                    nearest_start = node.query_block_start
            return nearest_start
    return nearest_start


def find_nearest_aligned_end(shortest_path_nodes, relative_end, relative_start):
    nearest_end = -1
    for i in range(len(shortest_path_nodes)):
        node = shortest_path_nodes[i]
        if relative_end <= node.query_block_end:
            if relative_end >= node.query_block_start:
                nearest_end = relative_end
            else:
                if i > 0 and shortest_path_nodes[i - 1].query_block_end > relative_start:
                    nearest_end = shortest_path_nodes[i - 1].query_block_end
            break
    if nearest_end == -1 and node.query_block_end < relative_end and node.query_block_end > relative_start:
        nearest_end = node.query_block_end
    return nearest_end


def convert_coord(relative_coord, shortest_path_nodes):
    lifted_coord = 0
    node_index = 0
    for i in range(0, len(shortest_path_nodes)):
        if relative_coord >= shortest_path_nodes[i].query_block_start:
            if relative_coord <= shortest_path_nodes[i].query_block_end:
                lifted_coord = (shortest_path_nodes[i].reference_block_start + (
                        relative_coord - shortest_path_nodes[i].query_block_start))
                node_index = i
    return lifted_coord, node_index


def find_deletions(start, end, shortest_path_nodes):
    deletions = 0
    for i in range(start + 1, end + 1):
        to_node = shortest_path_nodes[i]
        from_node = shortest_path_nodes[i - 1]
        deletions += (to_node.query_block_start - from_node.query_block_end - 1)
    return deletions


def find_insertions(start, end, shortest_path_nodes):
    insertions = 0
    for i in range(start + 1, end + 1):
        to_node = shortest_path_nodes[i]
        from_node = shortest_path_nodes[i - 1]
        insertions += (to_node.reference_block_start - from_node.reference_block_end - 1)
    return insertions


def find_mismatched_bases(start, end, shortest_path_nodes, parent):
    relative_coord1 = liftoff_utils.get_relative_child_coord(parent, start, shortest_path_nodes[0].is_reverse)
    relative_coord2 = liftoff_utils.get_relative_child_coord(parent, end, shortest_path_nodes[0].is_reverse)
    relative_start = min(relative_coord1, relative_coord2)
    relative_end = max(relative_coord1, relative_coord2)
    total_mismatches = 0
    for node in shortest_path_nodes:
        node_mismatches = np.array(node.mismatches)
        total_mismatches += len(
            node_mismatches[np.where((node_mismatches >= relative_start) & (node_mismatches <= relative_end))[0]])
    return total_mismatches



from multiprocessing import Pool
import math
from functools import partial
import numpy as np
from pyfaidx import Fasta, Faidx
import subprocess
import pysam
from lifton.liftoff  import aligned_seg, liftoff_utils
from os import path


def align_features_to_target(ref_chroms, target_chroms, args, feature_hierarchy, liftover_type, unmapped_features):
    if args.subcommand == "polish":
        sam_files = [args.directory + "/polish.sam"]
    else:
        target_fasta_dict = split_target_sequence(target_chroms, args.target, args.directory)
        genome_size = get_genome_size(target_fasta_dict)
        threads_per_alignment = max(1, math.floor(int(args.threads) / len(ref_chroms)))
        sam_files = []
        pool = Pool(int(args.threads))
        print("aligning features")
        func = partial(align_single_chroms, ref_chroms, target_chroms, threads_per_alignment, args, genome_size,
                       liftover_type)
        for result in pool.imap_unordered(func, np.arange(0, len(target_chroms))):
            sam_files.append(result)
        pool.close()
        pool.join()
    return parse_all_sam_files(feature_hierarchy, unmapped_features, liftover_type, sam_files)


def split_target_sequence(target_chroms, target_fasta_name, inter_files):
    Faidx(target_fasta_name)
    target_fasta_dict = Fasta(target_fasta_name, key_function=lambda x: x.split()[0])
    for chrm in target_chroms:
        if chrm != target_fasta_name:
            out = open(inter_files + "/" + chrm + ".fa", 'w')
            out.write(">" + chrm + "\n" + str(target_fasta_dict[chrm]))
    return target_fasta_dict


def get_genome_size(target_fasta_dict):
    genome_size = 0
    for value in target_fasta_dict.values():
        genome_size += len(value)
    return genome_size


def align_single_chroms(ref_chroms, target_chroms, threads, args, genome_size, liftover_type, index):
    max_single_index_size = 4000000000
    features_file, features_name = get_features_file(ref_chroms, args, liftover_type, index)
    target_file, output_file = get_target_file_and_output_file(liftover_type, target_chroms, index, features_name, args)
    threads_arg = str(threads)
    minimap2_path = get_minimap_path(args)
    target_prefix = get_target_prefix_name(target_chroms, index, args, liftover_type)
    if genome_size > max_single_index_size:
        split_prefix = args.directory + "/" + features_name + "_to_" + target_prefix + "_split"
        command = [minimap2_path, '-o', output_file, target_file, features_file] + args.mm2_options.split(" ") + [
            "--split-prefix", split_prefix, '-t', threads_arg]
        subprocess.run(command)
    else:
        minimap2_index = build_minimap2_index(target_file, args, threads_arg, minimap2_path)
        command = [minimap2_path, '-o', output_file, minimap2_index, features_file] + args.mm2_options.split(" ") + [
            '-t', threads_arg]
        subprocess.run(command)
    return output_file


def get_features_file(ref_chroms, args, liftover_type, index):
    if ref_chroms[index] == args.reference and (liftover_type == "chrm_by_chrm" or liftover_type == "copies"):
        features_name = 'reference_all'
    elif liftover_type == "unmapped":
        features_name = "unmapped_to_expected_chrom"
    elif liftover_type == "unplaced":
        features_name = "unplaced"
    else:
        features_name = ref_chroms[index]
    return args.directory + "/" + features_name + "_genes.fa", features_name


def get_target_file_and_output_file(liftover_type, target_chroms, index, features_name, args):
    if liftover_type != "chrm_by_chrm" or target_chroms[0] == args.target:
        target_file = args.target
        out_file_target = "target_all"
    else:
        target_file = args.directory + "/" + target_chroms[index] + ".fa"
        out_file_target = target_chroms[index]
    output_file = args.directory + "/" + features_name + "_to_" + out_file_target + ".sam"
    return target_file, output_file


def get_minimap_path(args):
    if args.m is None:
        minimap2 = "minimap2"
    else:
        minimap2 = args.m
    return minimap2


def get_target_prefix_name(target_chroms, index, args, liftover_type):
    if liftover_type != "chrm_by_chrm" or target_chroms[0] == args.target:
        prefix = "target_all"
    else:
        prefix = target_chroms[index]
    return prefix


def build_minimap2_index(target_file, args, threads, minimap2_path):
    if path.exists(target_file + ".mmi") is False:
        subprocess.run(
            [minimap2_path, '-d', target_file + ".mmi", target_file] + args.mm2_options.split(" ") + ['-t',
             threads ])
    return target_file + ".mmi"


def parse_all_sam_files(feature_hierarchy, unmapped_features, liftover_type, sam_files):
    aligned_segments_dict = {}
    for file in sam_files:
        aligned_segments = parse_alignment(file, feature_hierarchy, unmapped_features, liftover_type)
        aligned_segments_dict.update(aligned_segments)
    return aligned_segments_dict


def parse_alignment(file, feature_hierarchy, unmapped_features, search_type):
    all_aligned_blocks = {}
    sam_file = pysam.AlignmentFile(file, 'r', check_sq=False, check_header=False)
    sam_file_iter = sam_file.fetch()
    aln_id = 0
    name_dict = {}
    align_count_dict = {}
    for ref_seq in sam_file_iter:
        if ref_seq.is_unmapped is False:
            aln_id = add_alignment(ref_seq,  align_count_dict, search_type, name_dict,aln_id, feature_hierarchy,
                                   all_aligned_blocks)
        else:
            unmapped_features.append(feature_hierarchy.parents[ref_seq.query_name])
    remove_alignments_without_children(all_aligned_blocks, unmapped_features, feature_hierarchy)
    return all_aligned_blocks


def add_alignment(ref_seq, align_count_dict, search_type, name_dict, aln_id, feature_hierarchy,
                  all_aligned_blocks):
    ref_seq.query_name = edit_name(search_type, ref_seq, name_dict)
    aln_id += 1
    if ref_seq.query_name in align_count_dict:
        align_count = align_count_dict[ref_seq.query_name] + 1
    else:
        align_count = 0
    align_count_dict[ref_seq.query_name] = align_count
    aligned_blocks = get_aligned_blocks(ref_seq, aln_id, feature_hierarchy, search_type)
    if ref_seq.query_name in all_aligned_blocks:
        all_aligned_blocks[ref_seq.query_name].extend(aligned_blocks)
    else:
        all_aligned_blocks[ref_seq.query_name] = aligned_blocks
    return aln_id


def edit_name(search_type, ref_seq, name_dict):
    if search_type != "copies":
        return ref_seq.query_name + "_0"
    else:
        if ref_seq.query_name not in name_dict:
            name_dict[ref_seq.query_name] = 0
        name_dict[ref_seq.query_name] += 1
        return ref_seq.query_name + "_" + str(name_dict[ref_seq.query_name])


def get_aligned_blocks(alignment, aln_id, feature_hierarchy, search_type):
    cigar_operations = get_cigar_operations()
    cigar = alignment.cigar
    parent = feature_hierarchy.parents[liftoff_utils.convert_id_to_original(alignment.query_name)]
    query_start, query_end = get_query_start_and_end(alignment, cigar, cigar_operations)
    children = feature_hierarchy.children[liftoff_utils.convert_id_to_original(alignment.query_name)]
    end_to_end = is_end_to_end_alignment(parent, query_start, query_end)
    if search_type == "copies" and end_to_end is False:
        return []
    reference_block_start, reference_block_pos = alignment.reference_start, alignment.reference_start
    query_block_start, query_block_pos = query_start, query_start
    new_blocks, mismatches = [], []
    merged_children_coords = liftoff_utils.merge_children_intervals(children)
    for operation, length in cigar:
        if base_is_aligned(operation, cigar_operations):
            query_block_pos, reference_block_pos = add_aligned_base(operation, query_block_pos, reference_block_pos,
                                                                    length, cigar_operations, mismatches)
            if query_block_pos == query_end:
                add_block(query_block_pos, reference_block_pos, aln_id, alignment, query_block_start,
                          reference_block_start, mismatches, new_blocks, merged_children_coords, parent)
                break
        elif is_alignment_gap(operation, cigar_operations):
            add_block(query_block_pos, reference_block_pos, aln_id, alignment, query_block_start, reference_block_start,
                      mismatches, new_blocks, merged_children_coords, parent)
            mismatches, query_block_start, reference_block_start, query_block_pos, reference_block_pos = \
                end_block_at_gap(
                    operation, query_block_pos, reference_block_pos, length, cigar_operations)
    return new_blocks


def get_cigar_operations():
    return {"insertion": 1, "deletion": 2, "hard_clip": 5, "match": 7, "mismatch": 8}


def get_query_start_and_end(alignment, cigar, cigar_operations):
    query_start = alignment.query_alignment_start
    query_end = alignment.query_alignment_end
    if cigar[0][0] == cigar_operations["hard_clip"]:
        query_start += cigar[0][1]
        query_end += cigar[0][1]
    return query_start, query_end


def is_end_to_end_alignment(parent, query_start, query_end):
    return parent.end - parent.start + 1 == query_end - query_start


def base_is_aligned(operation, cigar_operations):
    return operation == cigar_operations["match"] or operation == cigar_operations["mismatch"]


def add_aligned_base(operation, query_block_pos, reference_block_pos, length, cigar_operations, mismatches):
    if operation == cigar_operations["mismatch"]:
        for i in range(query_block_pos, query_block_pos + length):
            mismatches.append(i)
    query_block_pos, reference_block_pos = adjust_position(operation, query_block_pos, reference_block_pos,
                                                           length, cigar_operations)
    return query_block_pos, reference_block_pos


def adjust_position(operation, query_block_pos, reference_block_pos, length, cigar_operations):
    if operation == cigar_operations["match"] or operation == cigar_operations["mismatch"] or operation == \
            cigar_operations["insertion"]:
        query_block_pos += length
    if operation == cigar_operations["match"] or operation == cigar_operations["mismatch"] or operation == \
            cigar_operations["deletion"]:
        reference_block_pos += length
    return query_block_pos, reference_block_pos


def add_block(query_block_pos, reference_block_pos, aln_id, alignment, query_block_start, reference_block_start,
              mismatches, new_blocks, merged_children_coords, parent):
    query_block_end = query_block_pos - 1
    reference_block_end = reference_block_pos - 1
    new_block = aligned_seg.aligned_seg(aln_id, alignment.query_name, alignment.reference_name, query_block_start,
                                        query_block_end,
                                        reference_block_start, reference_block_end, alignment.is_reverse,
                                        np.array(mismatches).astype(int))
    overlapping_children = find_overlapping_children(new_block, merged_children_coords, parent)
    if overlapping_children != []:

        new_blocks.append(new_block)




def find_overlapping_children(aln, children_coords, parent):
    overlapping_children = []
    for child_interval in children_coords:
        relative_start = liftoff_utils.get_relative_child_coord(parent, child_interval[0], aln.is_reverse)
        relative_end = liftoff_utils.get_relative_child_coord(parent, child_interval[1], aln.is_reverse)
        child_start, child_end = min(relative_start, relative_end), max(relative_start, relative_end)
        overlap = liftoff_utils.count_overlap(child_start, child_end, aln.query_block_start, aln.query_block_end)
        if overlap > 0:
            overlapping_children.append(child_start)
            overlapping_children.append(child_end)
    return overlapping_children


def is_alignment_gap(operation, cigar_operations):
    return operation == cigar_operations["insertion"] or operation == cigar_operations["deletion"]


def end_block_at_gap(operation, query_block_pos, reference_block_pos, length, cigar_operations):
    mismatches = []
    query_block_pos, reference_block_pos = adjust_position(operation, query_block_pos, reference_block_pos,
                                                           length, cigar_operations)
    query_block_start = query_block_pos
    reference_block_start = reference_block_pos
    return mismatches, query_block_start, reference_block_start, query_block_pos, reference_block_pos


def remove_alignments_without_children(all_aligned_blocks, unmapped_features, feature_hierarchy):
    features_to_remove = []
    for seq in all_aligned_blocks:
        if all_aligned_blocks[seq] == []:
            features_to_remove.append(seq)
            unmapped_features.append(feature_hierarchy.parents[liftoff_utils.convert_id_to_original(seq)])
    for feature in features_to_remove:
        del all_aligned_blocks[feature]
    return all_aligned_blocks



from lifton.liftoff  import liftoff_utils, new_feature


def merge_lifted_features(mapped_children, parent, unmapped_features, aln_cov_threshold, copy_id, feature_order,
                          feature_hierarchy, aln_cov, seq_id, seq_id_threshold):
    feature_list, final_features = {}, []
    non_parents = []
    top_target_feature = None
    if len(mapped_children) == 0:
        unmapped_features.append(parent)
        return []
    for child_name in mapped_children:
        child_feature = mapped_children[child_name]
        feature_list[child_feature.id] = child_feature
        if is_top_parent(child_feature, parent) is False:
            non_parents.append((child_feature, child_feature.attributes["Parent"][0]))
        else:
            top_target_feature = child_feature
    while (len(non_parents) != 0):
        non_parents, top_target_feature = create_parents(non_parents, parent, feature_hierarchy, feature_list)
    final_features = process_final_features_list(feature_list, top_target_feature, seq_id, seq_id_threshold, aln_cov,
                                                 aln_cov_threshold,
                                                 unmapped_features, parent, feature_order, copy_id)
    return final_features


def is_top_parent(child, parent):
    return child.id == parent.id


def create_parents(non_parents, top_ref_parent, feature_hierarchy, feature_list):
    added_parent_ids, new_non_parents = [], []
    top_target_feature = None
    for child, parent in non_parents:
        if parent not in added_parent_ids:
            target_parent_feature = make_new_parent(feature_list, parent, feature_hierarchy)
            if is_top_parent(target_parent_feature, top_ref_parent) is False:
                new_non_parents.append((target_parent_feature, target_parent_feature.attributes["Parent"][0]))
            else:
                top_target_feature = target_parent_feature
            added_parent_ids.append(parent)
    return new_non_parents, top_target_feature


def make_new_parent(feature_list, parent, feature_hierarchy):
    children = [feature for feature in feature_list.values() if
                "Parent" in feature.attributes and feature.attributes["Parent"][0] == parent]
    starts, ends = [child.start for child in children], [child.end for child in children]
    ref_parent = get_ref_parent(parent, feature_hierarchy)
    target_parent_feature = new_feature.new_feature(ref_parent.id, ref_parent.featuretype, children[0].seqid,
                                                    'Liftoff', children[0].strand, min(starts), max(ends),
                                                    ref_parent.frame, dict(ref_parent.attributes))
    feature_list[target_parent_feature.id] = target_parent_feature
    return target_parent_feature


def get_ref_parent(parent, feature_hierarchy):
    if parent in feature_hierarchy.parents:
        ref_parent = feature_hierarchy.parents[parent]
    else:
        ref_parent = feature_hierarchy.intermediates[parent]
    return ref_parent


def process_final_features_list(feature_list, top_target_feature, seq_id, seq_id_threshold, aln_cov, aln_cov_threshold,
                                unmapped_features, parent, feature_order, copy_id):
    final_features = [feature for feature in feature_list.values() if feature != top_target_feature]
    final_features.sort(key=lambda x: (x.seqid, x.start))
    final_features.sort(key=lambda x: feature_order[x.featuretype])
    final_features.insert(0, top_target_feature)
    if aln_cov < aln_cov_threshold or seq_id < seq_id_threshold:
        final_features = []
        unmapped_features.append(parent)
    else:
        top_target_feature.score = 1 - seq_id
        top_target_feature.attributes["copy_num_ID"] = [copy_id]
        top_target_feature.attributes["coverage"] = [str(aln_cov)[0:5]]
        top_target_feature.attributes["sequence_ID"] = [str(seq_id)[0:5]]
    return final_features



from lifton import get_id_fraction, lifton_class, logger
import math

def get_protein_boundary(cdss_aln_boundary, c_idx_last, c_idx, DEBUG):
    aa_start = cdss_aln_boundary[c_idx_last][0]
    aa_end = cdss_aln_boundary[c_idx-1][1]
    return aa_start, aa_end


def get_protein_reference_length_single(lifton_aln, c_idx, DEBUG):
    aa_start = 0
    aa_end = lifton_aln.cdss_protein_aln_boundaries[c_idx][1]
    aa_end = math.ceil(aa_end)
    ref_count = 0
    # logger.log(f"\t### protein chunk boundaries: {aa_start} - {aa_end}", debug=DEBUG)
    # logger.log("\tlifton_aln.ref_seq[aa_start:aa_start]: ", lifton_aln.ref_seq[aa_start:aa_end], debug=DEBUG)
    for i, letter in enumerate(lifton_aln.ref_seq[aa_start:aa_end]):
        if letter != "-":
            ref_count += 1
    return ref_count


def push_cds_idx(c_idx, lifton_aln, ref_aa_count, DEBUG):
    ref_aa_count = get_protein_reference_length_single(lifton_aln, c_idx, DEBUG)
    c_idx += 1
    return c_idx, ref_aa_count


def process_m_l_children(m_c_idx, m_c_idx_last, m_lifton_aln, l_c_idx, l_c_idx_last, l_lifton_aln, fai, chains, DEBUG):
    # logger.log(">> process_m_l_children!! ", debug=DEBUG)
    # logger.log("\tl_c_idx: ", l_c_idx, debug=DEBUG)
    # logger.log("\tm_c_idx: ", m_c_idx, debug=DEBUG)
    # logger.log("\tl_c_idx_last: ", l_c_idx_last, debug=DEBUG)
    # logger.log("\tm_c_idx_last: ", m_c_idx_last, debug=DEBUG)
    # logger.log("## m_cdss_aln_boundary: ", m_lifton_aln.cdss_protein_aln_boundaries, debug=DEBUG)
    # logger.log("## l_cdss_aln_boundary: ", l_lifton_aln.cdss_protein_aln_boundaries, debug=DEBUG)
    m_aa_start, m_aa_end = get_protein_boundary(m_lifton_aln.cdss_protein_aln_boundaries, m_c_idx_last, m_c_idx, DEBUG)
    l_aa_start, l_aa_end = get_protein_boundary(l_lifton_aln.cdss_protein_aln_boundaries, l_c_idx_last, l_c_idx, DEBUG)
    m_matches, m_length = get_id_fraction.get_partial_id_fraction(m_lifton_aln.ref_aln, m_lifton_aln.query_aln, math.floor(m_aa_start), math.ceil(m_aa_end))
    l_matches, l_length = get_id_fraction.get_partial_id_fraction(l_lifton_aln.ref_aln, l_lifton_aln.query_aln, math.floor(l_aa_start), math.ceil(l_aa_end))
    # logger.log(f"\tminiprot: {m_aa_start} - {m_aa_end}", debug=DEBUG)
    # logger.log(f"\tliftoff : {l_aa_start} - {l_aa_end}", debug=DEBUG)
    # logger.log(f"\tm_matches: {m_matches};  m_length: {m_length}", debug=DEBUG)
    # logger.log(f"\tl_matches: {l_matches};  l_length: {l_length}", debug=DEBUG)
    if (m_matches/m_length <= l_matches/l_length):
        cds_ls = create_lifton_entries(m_c_idx, m_c_idx_last, m_lifton_aln, l_c_idx, l_c_idx_last, l_lifton_aln, fai, False)
        chains.append(f"liftoff[{l_aa_start}-{l_aa_end}]")
    elif (m_matches/m_length > l_matches/l_length):
        cds_ls = create_lifton_entries(m_c_idx, m_c_idx_last, m_lifton_aln, l_c_idx, l_c_idx_last, l_lifton_aln, fai, True)
        chains.append(f"miniprot[{m_aa_start}-{m_aa_end}]")
    return cds_ls


def create_lifton_entries(m_c_idx, m_c_idx_last, m_lifton_aln, l_c_idx, l_c_idx_last, l_lifton_aln, fai, miniprot_is_better):
    cds_list = []
    source_aln = m_lifton_aln if miniprot_is_better else l_lifton_aln
    for c_idx in range(m_c_idx_last, m_c_idx) if miniprot_is_better else range(l_c_idx_last, l_c_idx):
        if source_aln.db_entry.strand == "+":
            c_idx_fix = c_idx
        elif source_aln.db_entry.strand == "-":
            c_idx_fix = len(source_aln.cds_children) - c_idx - 1
        lifton_cds = source_aln.cds_children[c_idx_fix]
        lifton_cds.attributes = l_lifton_aln.cds_children[0].attributes
        cds_list.append(lifton_class.Lifton_CDS(lifton_cds))
    return cds_list


def chaining_algorithm(l_lifton_aln, m_lifton_aln, fai, DEBUG):
    l_children = l_lifton_aln.cds_children
    m_children = m_lifton_aln.cds_children
    m_c_idx = 0
    l_c_idx = 0
    m_c_idx_last = m_c_idx
    l_c_idx_last = l_c_idx
    cds_list = []
    ref_aa_liftoff_count = 0
    ref_aa_miniprot_count = 0
    chains = []
    while m_c_idx != (len(m_children)-1) or l_c_idx != (len(l_children)-1):
        if (m_c_idx == len(m_children)-1) and (l_c_idx < (len(l_children)-1)):
            l_c_idx, ref_aa_liftoff_count = push_cds_idx(l_c_idx, l_lifton_aln, ref_aa_liftoff_count, DEBUG)
            continue
        if (m_c_idx < len(m_children)-1) and (l_c_idx == (len(l_children)-1)):
            m_c_idx, ref_aa_miniprot_count = push_cds_idx(m_c_idx, m_lifton_aln, ref_aa_miniprot_count, DEBUG)
            continue
        # logger.log(f">> ref_aa_miniprot_count: {ref_aa_miniprot_count}", debug=DEBUG)
        # logger.log(f">> ref_aa_liftoff_count : {ref_aa_liftoff_count}", debug=DEBUG)
        # logger.log(f">> len(m_children)      : {len(m_children)}", debug=DEBUG)
        # logger.log(f">> len(l_children)      : {len(l_children)}", debug=DEBUG)
        if m_lifton_aln.db_entry.strand == "+":
            m_c = m_children[m_c_idx]
            l_c = l_children[l_c_idx]
            if ref_aa_liftoff_count < ref_aa_miniprot_count:
                l_c_idx, ref_aa_liftoff_count = push_cds_idx(l_c_idx, l_lifton_aln, ref_aa_liftoff_count, DEBUG)
            elif ref_aa_liftoff_count > ref_aa_miniprot_count:
                m_c_idx, ref_aa_miniprot_count = push_cds_idx(m_c_idx, m_lifton_aln, ref_aa_miniprot_count, DEBUG)
            elif ref_aa_liftoff_count == ref_aa_miniprot_count:
                # 1. Accumulated aa in the reference of Liftoff and miniprot are the same
                # 2. CDS ends at the same position.
                if l_c_idx > 0 and m_c_idx > 0 and m_c.end == l_c.end:
                    cdss = process_m_l_children(m_c_idx, m_c_idx_last, m_lifton_aln, l_c_idx, l_c_idx_last, l_lifton_aln, fai, chains, DEBUG)
                    cds_list += cdss
                    m_c_idx_last = m_c_idx
                    l_c_idx_last = l_c_idx
                l_c_idx, ref_aa_liftoff_count = push_cds_idx(l_c_idx, l_lifton_aln, ref_aa_liftoff_count, DEBUG)
                m_c_idx, ref_aa_miniprot_count = push_cds_idx(m_c_idx, m_lifton_aln, ref_aa_miniprot_count, DEBUG)
        elif m_lifton_aln.db_entry.strand == "-":
            m_c = m_children[len(m_children) - m_c_idx - 1]
            l_c = l_children[len(l_children) - l_c_idx - 1]
            if ref_aa_liftoff_count < ref_aa_miniprot_count:
                l_c_idx, ref_aa_liftoff_count = push_cds_idx(l_c_idx, l_lifton_aln, ref_aa_liftoff_count, DEBUG)
            elif ref_aa_liftoff_count > ref_aa_miniprot_count:
                m_c_idx, ref_aa_miniprot_count = push_cds_idx(m_c_idx, m_lifton_aln, ref_aa_miniprot_count, DEBUG)
            elif ref_aa_liftoff_count == ref_aa_miniprot_count:
                # 1. Accumulated aa in the reference of Liftoff and miniprot are the same
                # 2. CDS ends at the same position.
                if l_c_idx > 0 and m_c_idx > 0 and m_c.end == l_c.end:
                    cdss = process_m_l_children(m_c_idx, m_c_idx_last, m_lifton_aln, l_c_idx, l_c_idx_last, l_lifton_aln, fai, chains, DEBUG)
                    cds_list += cdss
                    m_c_idx_last = m_c_idx
                    l_c_idx_last = l_c_idx
                l_c_idx, ref_aa_liftoff_count = push_cds_idx(l_c_idx, l_lifton_aln, ref_aa_liftoff_count, DEBUG)
                m_c_idx, ref_aa_miniprot_count = push_cds_idx(m_c_idx, m_lifton_aln, ref_aa_miniprot_count, DEBUG)
    l_c_idx += 1
    m_c_idx += 1
    cdss = process_m_l_children(m_c_idx, m_c_idx_last, m_lifton_aln, l_c_idx, l_c_idx_last, l_lifton_aln, fai, chains, DEBUG)
    cds_list += cdss
    return cds_list, chains


from collections import defaultdict
import warnings
from Bio.Seq import Seq
from lifton.liftoff  import liftoff_utils
import parasail
from itertools import groupby
import numpy as np
import re


def polish_annotations(feature_list, ref_faidx, target_faidx, args, feature_heirarchy, target_feature):
    target_sub_features = get_sub_features(feature_list, target_feature)
    ref_sub_features = get_sub_features(feature_heirarchy.children, target_sub_features[0].id)
    ref_gene = feature_heirarchy.parents[target_sub_features[0].id]

    target_sub_features = feature_list[target_feature]
    transcript_list = [feature for feature in target_sub_features if "matches_ref_protein" in feature.attributes]
    good_transcripts = [tran for tran in transcript_list if  tran.attributes["valid_ORF"] == ["True"]]
    if len(transcript_list) != len(good_transcripts):
        output_sam = open(args.directory+"/polish.sam", 'w')
        write_sam_header(target_faidx, output_sam)
        target_gene = target_sub_features[0]
        polish_annotation(ref_gene, target_gene, ref_sub_features, target_sub_features, ref_faidx,
                          target_faidx,output_sam)
        return True
    return False


def get_sub_features(feature_list, feature_name):
    return feature_list[feature_name]


def find_and_check_cds(target_sub_features, ref_sub_features, ref_faidx, target_faidx, feature_list):
    cds_features = get_cds_features(target_sub_features)
    total_cds = 0
    num_good_cds = 0
    if len(cds_features) > 0:
        total_cds, num_good_cds = count_good_cds(cds_features, ref_faidx, target_faidx, ref_sub_features,
                                                 target_sub_features, feature_list)
        target_sub_features[0].attributes["valid_ORFs"] = [str(num_good_cds)]
    return total_cds, num_good_cds


def write_sam_header(target_fa, output_sam):
    for seq in target_fa.keys():
        output_sam.write("@SQ" + "\t" + "SN:" + seq + "\t" + "LN:" + str(len(target_fa[seq])) + "\n")


def get_cds_features(sub_features):
    return [feature for feature in sub_features if feature.featuretype == "CDS"]


def count_good_cds(cds_features, ref_faidx, target_faidx, ref_children, target_sub_features, feature_list):
    grouped_cds = group_cds_by_tran(cds_features)
    good_cds_count = 0
    for cds_group in grouped_cds:
        cds_seq = get_seq(cds_group, target_faidx)
        transcript = [feature for feature in target_sub_features if feature.id == cds_group[0].attributes["Parent"][
            0]][0]
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            protein = cds_seq.translate()
        if matches_ref(cds_group, ref_faidx, ref_children, protein):
            transcript.attributes["matches_ref_protein"] = ['True']
        else:
            transcript.attributes["matches_ref_protein"] = ['False']
        longest_ORF, longest_ORF_coords = get_longest_ORF(cds_seq)
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            protein = longest_ORF.translate()
        transcript.attributes["valid_ORF"] = ['False']
        if len(protein) < 3:
            transcript.attributes["partial_ORF"] = ['True']
        elif missing_start(protein):
                transcript.attributes["missing_start_codon"] = ['True']
        elif missing_stop(protein):
            transcript.attributes["missing_stop_codon"] = ['True']
        elif inframe_stop(protein):
            transcript.attributes["inframe_stop_codon"] = ['True']
        else:
            transcript.attributes["valid_ORF"] = ['True']
            good_cds_count +=1
        if longest_ORF != cds_seq:
            adjust_cds_coords(cds_group,longest_ORF_coords, feature_list )
    return len(grouped_cds), good_cds_count


def get_longest_ORF(cds_seq):
    longest_ORF_seq = cds_seq
    longest_ORF = [0, len(cds_seq)]
    ORFs = [[m.start(), m.end()] for m in re.finditer(r'ATG(?:(?!TAA|TAG|TGA)...)*(?:TAA|TAG|TGA)',
                                                      str(cds_seq))]
    ORFs.sort(key=lambda x: x[1] - x[0])
    if len(ORFs) >0 :
        possible_ORF = ORFs[-1]
        if  possible_ORF[1] - possible_ORF[0] != len(cds_seq) and possible_ORF[1] - possible_ORF[0] >= \
                180:
            longest_ORF_seq = cds_seq[longest_ORF[0]:longest_ORF[1]]
            longest_ORF = possible_ORF
    return longest_ORF_seq, longest_ORF


def adjust_cds_coords(cds_features, longest_ORF_coords, feature_list):
    coords = np.hstack(np.array([list(np.arange(cds.start, cds.end +1)) for cds in cds_features], dtype=object))
    for cds in cds_features:
        relative_start = np.where(coords==cds.start)[0][0]
        relative_end = np.where(coords == cds.end)[0][0]
        if relative_end < longest_ORF_coords[0]:
            feature_list.remove([item for item in feature_list if item.id == cds.id][0])
        if relative_start > longest_ORF_coords[1]:
            feature_list.remove([item for item in feature_list if item.id == cds.id][0])
        if longest_ORF_coords[0] >= relative_start and  longest_ORF_coords[0] <= relative_end:
            cds.start = cds.start + (longest_ORF_coords[0] - relative_start)
        if relative_start <= longest_ORF_coords[1] and relative_end >= longest_ORF_coords[1]:
            cds.end = cds.end - (relative_end - longest_ORF_coords[1] +1)


def group_cds_by_tran(cds_features):
    groups = defaultdict(list)
    for cds in cds_features:
        groups[cds.attributes["Parent"][0]].append(cds)
    return groups.values()



def get_seq(cds_group, fasta_index):
    cds_group.sort(key = lambda x: x.start)
    cds_seq = ""
    chrom_seq = fasta_index[cds_group[0].seqid]
    for cds in cds_group:
        sub_seq = chrom_seq[cds.start -1:cds.end]
        cds_seq += sub_seq.seq
    if cds.strand == "-":
        cds_seq = Seq(cds_seq).reverse_complement()
    else:
        cds_seq = Seq(cds_seq)
    return cds_seq.upper()


def missing_start(protein):
    return protein[0] != "M"


def missing_stop(protein):
    return "*" != protein[-1]


def inframe_stop(protein):
    return "*" in protein[:-1]


def matches_ref(cds_group, ref_fa, ref_children, target_protein):
    tran = cds_group[0].attributes["Parent"]
    ref_cds_group= [ref_cds for ref_cds in ref_children if ref_cds.featuretype=="CDS" and ref_cds.attributes[
        "Parent"] ==
                    tran]
    ref_cds_seq = get_seq(ref_cds_group, ref_fa)
    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        ref_protein = ref_cds_seq.translate()
    return target_protein == ref_protein


def polish_annotation(ref_gene, target_gene, ref_children, target_children, ref_fa, target_fa, output_sam):
    ref_exons = [feature for feature in ref_children if feature.featuretype == "exon"]
    target_exons = [feature for feature in target_children if feature.featuretype == "exon"]
    ref_CDS = [feature for feature in ref_children if feature.featuretype == "CDS"]
    if len(ref_exons) == 0:
        ref_exons = ref_CDS
        target_exons = [feature for feature in target_children if feature.featuretype == "CDS"]
    ref_CDS_intervals = liftoff_utils.merge_children_intervals(ref_CDS)
    splice_sites = add_splice_sites(ref_exons, ref_gene)
    merged_ref_intervals = liftoff_utils.merge_children_intervals(ref_exons)
    exon_group_dict = find_overlapping_exon_groups(merged_ref_intervals, ref_exons)
    matrix = make_scoring_matrix(3)
    for i in range (len(merged_ref_intervals)):
        ref_interval =  merged_ref_intervals[i]
        target_interval = get_target_interval(exon_group_dict[i], target_exons, ref_interval)
        flank = max(0, (ref_interval[1] - ref_interval[0])- (target_interval[1]- target_interval[0])) + 10
        if target_interval != [0,0]:
            is_reverse = ref_gene.strand != target_gene.strand
            ref_seq = get_feature_sequence(ref_interval, ref_fa, ref_gene.seqid,  0 )
            capitalized_ref_seq = cds_and_splice_sites_to_upper(ref_interval, splice_sites, ref_CDS_intervals,
                                                                ref_seq, is_reverse)
            target_seq = get_feature_sequence(target_interval, target_fa, target_gene.seqid, flank)
            alignment = parasail.sg_dx_trace_scan_sat(capitalized_ref_seq, target_seq, 10, 1, matrix)
            write_sam_file(ref_interval, target_interval, ref_gene, target_gene, capitalized_ref_seq, alignment,
                           output_sam, flank)
    remove_splice_sites(ref_exons, ref_gene)


def add_splice_sites(exons, parent):
    splice_sites = []
    for exon in exons:
        if exon.start - 2 >= parent.start:
            exon.start = exon.start - 2
            splice_sites.append([exon.start, exon.start + 1])
        if exon.end + 2 <= parent.end:
            exon.end = exon.end + 2
            splice_sites.append([exon.end - 1, exon.end])
    return splice_sites

def remove_splice_sites(exons, parent):
    for exon in exons:
        if exon.start  != parent.start:
            exon.start = exon.start + 2
        if exon.end  != parent.end:
            exon.end = exon.end - 2



def find_overlapping_exon_groups(merged_ref_intervals, ref_exons):
    exon_group_dict = {}
    for i in range(len(merged_ref_intervals)):
        exon_group_dict[i] = []
        interval = merged_ref_intervals[i]
        for exon in ref_exons:
            if interval[0] <= exon.start <= interval[1]:
                exon_group_dict[i].append(exon.id)
    return exon_group_dict


def make_scoring_matrix(match_reward_coding):
    match_reward_non_coding = 2
    mismatch_penalty = -4
    matrix = parasail.matrix_create("ACGTacgt*", match_reward_non_coding, mismatch_penalty, True)
    matrix[0,4] = match_reward_coding
    matrix[1,5] = match_reward_coding
    matrix[2,6] = match_reward_coding
    matrix[3,7] = match_reward_coding
    matrix[4,0] = match_reward_coding
    matrix[5,1] = match_reward_coding
    matrix[6,2] = match_reward_coding
    matrix[7,3] = match_reward_coding
    return matrix


def get_target_interval(exon_group, target_exons, ref_interval):
    target_group = []
    for ref_exon in exon_group:
        matching_target_exon = [exon for exon in target_exons if exon.id == ref_exon]
        if len(matching_target_exon) !=0:
            target_group.append(matching_target_exon[0])
    if len(target_group) != 0:
        start = min([exon.start for exon in target_group])
        end = max([exon.end for exon in target_group])
    else:
        start = min([exon.start for exon in target_exons]) - (ref_interval[1]-ref_interval[0])
        end = max([exon.end for exon in target_exons]) + (ref_interval[1]-ref_interval[0])
    return [start, end]


def get_feature_sequence(interval, fa, chrm,  flank):
    chrom_seq = fa[chrm]
    flank_start = get_flank_start(interval[0], flank)
    flank_end = get_flank_end(interval[1], flank, len(chrom_seq))
    seq = chrom_seq[flank_start - 1: flank_end]
    final_seq = seq.seq.lower()
    return final_seq


def get_flank_start(start, flank):
    return round(max(1, start - flank))


def get_flank_end(end, flank, chrom_length):
    return round(min(end + flank, chrom_length))


def cds_and_splice_sites_to_upper(ref_interval, splice_sites, cds_intervals, ref_seq, is_reverse):
    cds_and_splice_sites = splice_sites + cds_intervals
    cap_segs = [[start -ref_interval[0], end-ref_interval[0]] for start, end in cds_and_splice_sites if start >=
                ref_interval[0] and end <= ref_interval[1]]
    prev_string = ref_seq
    for start,end in cap_segs:
        new_string = prev_string[:start] + prev_string[start:end+1].upper() + prev_string[end+1:]
        prev_string = new_string
    if is_reverse:
        return str(Seq(prev_string).reverse_complement())
    return prev_string


def write_sam_file(ref_interval, target_interval, ref_gene, target_gene, ref_seq, alignment, output_sam, flank ):
    reference_traceback = alignment.traceback.query
    target_traceback = alignment.traceback.ref
    is_reverse = ref_gene.strand != target_gene.strand
    if is_reverse:
        hard_clip_start = ref_gene.end - ref_interval[1]
    else:
        hard_clip_start = ref_interval[0] - ref_gene.start
    cigar_string, target_start_offset = make_cigar(reference_traceback, target_traceback, hard_clip_start,)
    if is_reverse:
        bit_flag = 16
    else:
        bit_flag = 0
    target_start_with_flank = get_flank_start(target_interval[0], flank)
    sam_fields = np.array([target_gene.id, bit_flag, target_gene.seqid,
                           target_start_with_flank + int(
        target_start_offset),
                           0, cigar_string,"*", 0,0, ref_seq.upper(), "*"])

    output_sam.write('\t'.join(sam_fields.astype(str)) + "\n")


def make_cigar(reference_traceback, target_traceback, hard_clip_start):
    expanded_cigar = ""
    chars = [pos for pos, char in enumerate(reference_traceback) if char != "-"]
    start = chars[0]
    for i in range(start, len(reference_traceback)):
        if reference_traceback[i].upper() == target_traceback[i].upper():
            expanded_cigar+= '='
        elif reference_traceback[i] == "-":
            expanded_cigar += 'D'
        elif reference_traceback[i] != "-" and target_traceback[i]=="-":
            expanded_cigar += 'I'
        elif reference_traceback[i] != "-" and target_traceback[i]!="-":
            expanded_cigar += 'X'
        else:
            print(reference_traceback[i], target_traceback[i])
    cigar = condense_cigar_string(expanded_cigar, hard_clip_start)
    return cigar, start


def condense_cigar_string(expanded_cigar, hard_clip_start):
    cigar_numbers = ([[str(sum(1 for _ in g)), k] for k, g in groupby(expanded_cigar)])
    if cigar_numbers[0][1] == 'D':
        cigar_numbers = cigar_numbers[1:]
    cigar = ''.join([i for sub in cigar_numbers for i in sub])
    if hard_clip_start > 0:
        cigar = str(hard_clip_start) + 'H' + cigar
    return cigar



from lifton import align, logger, lifton_class, lifton_utils
import subprocess, os, sys, copy
from intervaltree import Interval, IntervalTree


def check_miniprot_installed():
    """
        This function checks if miniprot is installed.

        Parameters:
        None

        Returns:
        True if miniprot is installed, False otherwise.
    """
    miniprot_path = "miniprot"
    command = [miniprot_path, "--version"]
    installed = False
    try:
        res = subprocess.run(command)
        installed = True
    except: 
        pass
    return installed


def run_miniprot(outdir, args, tgt_genome, ref_proteins_file):
    """
        This function runs miniprot.

        Parameters:
        - outdir: output directory
        - args: arguments
        - tgt_genome: target genome
        - ref_proteins_file: reference protein file

        Returns:
        miniprot_output: miniprot output file
    """
    miniprot_outdir = outdir + "miniprot/"
    os.makedirs(miniprot_outdir, exist_ok=True)
    miniprot_output = miniprot_outdir + "miniprot.gff3"
    miniprot_path = "miniprot"
    print("args.mp_options: ", args.mp_options)
    command = [miniprot_path, "--gff-only", tgt_genome, ref_proteins_file] + args.mp_options.split(" ")
    print("miniprot: ", " ".join(command))
    try:
        fw = open(miniprot_output, "w")
        subprocess.run(command, stdout=fw)
        fw.close()
    except: 
        print("failed to run miniprot")
        sys.exit(1)
    return miniprot_output


def lifton_miniprot_with_ref_protein(m_feature, m_feature_db, ref_db, ref_gene_id, ref_trans_id, tgt_fai, 
ref_proteins, ref_trans, tree_dict, ref_features_dict, args):
    """
        This function create a miniprot gene entry with reference protein.

        Parameters:
        - m_feature: miniprot feature
        - m_feature_db: miniprot feature database
        - ref_db: reference database
        - ref_gene_id: reference gene ID
        - ref_trans_id: reference transcript ID
        - tgt_fai: target fasta index
        - ref_proteins: reference protein dictionary
        - ref_trans: reference transcript dictionary
        - tree_dict: tree dictionary
        - ref_features_dict: reference features dictionary
        - args: LiftOn arguments

        Returns:
        lifton_gene: LiftOn gene instance
        lifton_transcript_id: LiftOn transcript ID
        lifton_status: LiftOn status
    """
    mtrans_id = m_feature.attributes["ID"][0]
    # Create LifOn gene instance
    m_gene_feature = copy.deepcopy(m_feature)
    m_gene_feature.featuretype = "gene"
    lifton_gene = lifton_class.Lifton_GENE(ref_gene_id, m_gene_feature, copy.deepcopy(ref_db[ref_gene_id].attributes), tree_dict, ref_features_dict, args)
    lifton_gene.update_gene_info(m_feature.seqid, m_feature.start, m_feature.end)
    # Create LifOn transcript instance
    Lifton_trans = lifton_gene.add_miniprot_transcript(ref_trans_id, copy.deepcopy(m_feature), ref_db[ref_trans_id].attributes, ref_features_dict)
    lifton_gene.update_trans_info(Lifton_trans.entry.id, m_feature.seqid, m_feature.start, m_feature.end)
    # Create exon / CDS entries
    cdss = m_feature_db.children(m_feature, featuretype='CDS')  # Replace 'exon' with the desired child feature type
    for cds in list(cdss):
        lifton_gene.add_exon(Lifton_trans.entry.id, cds)
        cds_copy = copy.deepcopy(cds)
        lifton_gene.add_cds(Lifton_trans.entry.id, cds_copy)
    # Update LiftOn status
    lifton_status = lifton_class.Lifton_Status()                
    m_entry = m_feature_db[mtrans_id]
    m_lifton_aln = align.lifton_parasail_align(Lifton_trans, m_entry, tgt_fai, ref_proteins, ref_trans_id)
    lifton_status.annotation =  "miniprot"
    lifton_status.lifton_aa = m_lifton_aln.identity
    return lifton_gene, Lifton_trans, Lifton_trans.entry.id, lifton_status


def process_miniprot(mtrans, ref_db, m_feature_db, tree_dict, tgt_fai, ref_proteins, ref_trans, ref_features_dict, fw_score, m_id_2_ref_id_trans_dict, ref_features_len_dict, ref_trans_exon_num_dict, ref_features_reverse_dict, args):
    mtrans_id = mtrans.attributes["ID"][0]
    mtrans_interval = Interval(mtrans.start, mtrans.end, mtrans_id)
    is_overlapped = lifton_utils.check_ovps_ratio(mtrans, mtrans_interval, args.overlap, tree_dict)
    lifton_gene = None
    lifton_trans = None
    if not is_overlapped:
        ref_trans_id = m_id_2_ref_id_trans_dict[mtrans_id]            
        ref_gene_id, ref_trans_id = lifton_utils.get_ref_ids_miniprot(ref_features_reverse_dict, mtrans_id, m_id_2_ref_id_trans_dict)
        if ref_gene_id is None:
            return None
        if ref_trans_id in ref_proteins.keys() and ref_trans_id in ref_trans.keys():
            if ref_trans_id != None:
                # Check if the additional copy is valid
                # 1. Remove processed pseudogenes: 1 CDS in miniprot but >1 CDS in reference
                # 2. Check the trans ratio is in the range of min_minprot and max_minprot
                if len(list(m_feature_db.children(mtrans, featuretype='CDS'))) == 1 and ref_trans_exon_num_dict[ref_trans_id] > 1: # Processed pseudogene
                    return None
                miniprot_trans_ratio = (mtrans.end - mtrans.start + 1) / ref_features_len_dict[ref_gene_id]
                if miniprot_trans_ratio > args.min_miniprot and miniprot_trans_ratio < args.max_miniprot:
                    lifton_gene, lifton_trans, transcript_id, lifton_status = lifton_miniprot_with_ref_protein(mtrans, m_feature_db, ref_db.db_connection, ref_gene_id, ref_trans_id, tgt_fai, ref_proteins, ref_trans, tree_dict, ref_features_dict, args)
                    lifton_gene.transcripts[transcript_id].entry.attributes["miniprot_annotation_ratio"] = [f"{miniprot_trans_ratio:.3f}"]
                else: # Invalid miniprot transcript
                    return None
            else: # Skip those cannot be found in reference.
                return None
        lifton_trans_aln, lifton_aa_aln = lifton_gene.orf_search_protein(lifton_trans.entry.id, ref_trans_id, tgt_fai, ref_proteins, ref_trans, lifton_status)
        lifton_utils.print_lifton_status(transcript_id, mtrans, lifton_status, DEBUG=args.debug)
        lifton_gene.add_lifton_gene_status_attrs("miniprot")
        lifton_gene.add_lifton_trans_status_attrs(transcript_id, lifton_status)
        lifton_utils.write_lifton_status(fw_score, transcript_id, mtrans, lifton_status)
    return lifton_gene


from lifton import intervals, lifton_utils, annotation, extract_sequence, stats, logger, run_liftoff, run_miniprot, run_evaluation, __version__
from intervaltree import Interval
import argparse
from pyfaidx import Fasta
import os, sys
import time

def args_gffutils(parser):
    gffutils_grp = parser.add_argument_group('* gffutils parameters')
    gffutils_grp.add_argument(
        '--merge-strategy', default='create_unique', choices=['create_unique', 'merge', 'error', 'warning', 'replace'],
        help='Strategy for merging features when building the database. Default is "create_unique".'
    )
    gffutils_grp.add_argument(
        '--id-spec', default=None,
        help='Attribute to use as feature ID. Default is "ID".'
    )
    gffutils_grp.add_argument(
        '--force', default=False, action='store_true', help='Overwrite existing database.'
    )
    gffutils_grp.add_argument(
        '--verbose', default=False, action='store_true', help='Enable verbose output.'
    )
    return gffutils_grp


def args_outgrp(parser):
    outgrp = parser.add_argument_group('* Output settings')
    outgrp.add_argument(
        '-o', '--output', default='lifton.gff3', metavar='FILE',
        help='write output to FILE in same format as input; by default, output is written to "lifton.gff3"'
    )
    outgrp.add_argument(
        '-u', default='unmapped_features.txt', metavar='FILE',
        help='write unmapped features to FILE; default is "unmapped_features.txt"',
    )
    outgrp.add_argument(
        '-exclude_partial', action='store_true',
        help='write partial mappings below -s and -a threshold to unmapped_features.txt; if true '
             'partial/low sequence identity mappings will be included in the gff file with '
             'partial_mapping=True, low_identity=True in comments'
    )
    return outgrp


def args_aligngrp(parser):
    aligngrp = parser.add_argument_group('Alignments')
    aligngrp.add_argument('-mm2_options', metavar='=STR', type=str, default='-a --end-bonus '
                                                                            '5 --eqx -N 50 '
                                                                            '-p 0.5',
                          help='space delimited minimap2 parameters. By default ="-a --end-bonus 5 --eqx -N 50 -p 0.5"')
    aligngrp.add_argument('-mp_options', metavar='=STR', type=str, default='',
                          help='space delimited miniprot parameters.')
    aligngrp.add_argument(
        '-a', default=0.5, metavar='A', type=float,
        help='designate a feature mapped only if it aligns with coverage ≥A; by default A=0.5',
    )
    aligngrp.add_argument(
        '-s', default=0.5, metavar='S', type=float,
        help='designate a feature mapped only if its child features (usually exons/CDS) align '
             'with sequence identity ≥S; by default S=0.5'
    )
    aligngrp.add_argument(
        '-min_miniprot', default=0.9, metavar='MIN_MINIPROT', type=float,
        help='The minimum length ratio of a protein-coding transcript to the longest protein-coding transcript within a gene locus, as identified exclusively by miniprot in the target genome, is set by default to MIN_MINIPROT=0.9.'
    )
    aligngrp.add_argument(
        '-max_miniprot', default=1.5, metavar='MAX_MINIPROT', type=float,
        help='The maximum length ratio of a protein-coding transcript to the longest protein-coding transcript within a gene locus, as identified exclusively by miniprot in the target genome, is set by default to MIN_MINIPROT=1.5.'
    )
    aligngrp.add_argument(
        '-d', metavar='D', default=2.0, type=float,
        help='distance scaling factor; alignment nodes separated by more than a factor of D in '
             'the target genome will not be connected in the graph; by default D=2.0'
    )
    aligngrp.add_argument(
        '-flank', default=0, metavar='F', type=float, help="amount of flanking sequence to align as a "
                                                           "fraction [0.0-1.0] of gene length. This can improve gene "
                                                           "alignment where gene structure  differs between "
                                                           "target and "
                                                           "reference; by default F=0.0")
    return aligngrp


def args_optional(parser):
    parser.add_argument('-V', '--version', help='show program version', action='version', version=__version__)
    parser.add_argument('-D', '--debug', action='store_true', help='Run debug mode')
    parser.add_argument(
        '-t', '--threads', default=1, type=int, metavar='THREADS', help='use t parallel processes to accelerate alignment; by default p=1'
    )
    parser.add_argument('-m', help='Minimap2 path', metavar='PATH')
    parser.add_argument('-f', '--features', metavar='TYPES', help='list of feature types to lift over')
    parser.add_argument(
        '-infer-genes', required=False, action='store_true',
        help='use if annotation file only includes transcripts, exon/CDS features. '
             'Automatically enabled for GTF files.'
    )
    parser.add_argument(
        '-infer_transcripts', action='store_true', required=False,
        help='use if annotation file only includes exon/CDS features and does not include transcripts/mRNA. '
             'Automatically enabled for GTF files.'
    )
    parser.add_argument(
        '-chroms', metavar='TXT', help='comma seperated file with corresponding chromosomes in '
                                       'the reference,target sequences',
    )
    parser.add_argument(
        '-unplaced', metavar='TXT',
        help='text file with name(s) of unplaced sequences to map genes from after genes from '
             'chromosomes in chroms.txt are mapped; default is "unplaced_seq_names.txt"',
    )
    parser.add_argument('-copies', action='store_true', help='look for extra gene copies in the target genome')
    parser.add_argument(
        '-sc', default=1.0, metavar='SC', type=float,
        help='with -copies, minimum sequence identity in exons/CDS for which a gene is considered '
             'a copy; must be greater than -s; default is 1.0',
    )
    parser.add_argument('-overlap', default=0.1, metavar='O', help="maximum fraction [0.0-1.0] of overlap allowed by 2 "
                                                                   "features; by default O=0.1", type=float)
    parser.add_argument('-mismatch', default=2, metavar='M', help="mismatch penalty in exons when finding best "
                                                                  "mapping; by default M=2", type=int)
    parser.add_argument('-gap_open', default=2, metavar='GO', help="gap open penalty in exons when finding best "
                                                                   "mapping; by default GO=2", type=int)
    parser.add_argument('-gap_extend', default=1, metavar='GE', help="gap extend penalty in exons when finding best "
                                                                     "mapping; by default GE=1", type=int)
    parser.add_argument('-subcommand', required=False,  help=argparse.SUPPRESS)
    parser.add_argument('-polish', required=False, action='store_true', default = False)
    parser.add_argument('-cds', required=False, action="store_true", default=True, help="annotate status of each CDS "
                                                                                        "(partial, missing start, "
                                                                                        "missing stop, inframe stop "
                                                                                        "codon)")
    parser.add_argument(
        '-time', '--measure_time', required=False, action='store_true',
        help='Enable time measurement for each step'
    )


def parse_args(arglist):
    parser = argparse.ArgumentParser(description='Lift features from one genome assembly to another')
    parser.add_argument('target', help='target fasta genome to lift genes to')
    parser.add_argument('reference', help='reference fasta genome to lift genes from')
    parser.add_argument('-E', '--evaluation', help='Run LiftOn in evaluation mode', action='store_true', default = False)
    parser.add_argument('-EL', '--evaluation-liftoff-chm13', help='Run LiftOn in evaluation mode', action='store_true', default = False)
    parser.add_argument('-c', '--write_chains', help='Write chaining files', action='store_true', default = True)
    parser.add_argument('--no-orf-search', help='Do not perform open reading frame search', action='store_true', default = False)
    parser_outgrp = args_outgrp(parser)
    parser_aligngrp = args_aligngrp(parser)
    args_optional(parser)
    referencegrp = parser.add_argument_group('* Required input (Reference annotation)')    
    referencegrp.add_argument(
        '-g', '--reference-annotation', metavar='GFF',  required=True,
        help='the reference annotation file to lift over in GFF3 or GTF format (or) '
                'name of feature database. GTF files are automatically detected and converted to GFF3 for better compatibility. '
                'For best results with GTF files, ensure agat or gffread is installed.'
    )
    ###################################
    # START for the LiftOn params
    ###################################
    referenceseqgrp = parser.add_argument_group('* Optional input (Reference sequences)')
    referenceseqgrp.add_argument(
        '-P', '--proteins', metavar='FASTA', required=False, default=None,
        help='the reference protein sequences. If not provided, the protein sequences will be extracted from the reference annotation. The ID of the protein sequences must match the ID of the transcript sequences.'
    )
    referenceseqgrp.add_argument(
        '-T', '--transcripts', metavar='FASTA', required=False, default=None,
        help='the reference transcript sequences. If not provided, the transcript sequences will be extracted from the reference annotation.'
    )
    liftoffrefrgrp = parser.add_argument_group('* Optional input (Liftoff annotation)')
    liftoffrefrgrp.add_argument(
        '-L', '--liftoff', metavar='gff', default=None,
        help='the annotation generated by Liftoff (or) '
                'name of Liftoff gffutils database; if not specified, '
                'a Liftoff database will be built automatically'
    )
    miniprotrefrgrp = parser.add_argument_group('* Optional input (miniprot annotation)')
    miniprotrefrgrp.add_argument(
        '-M', '--miniprot', metavar='gff', default=None,
        help='the annotation generated by miniprot (or) '
                'name of miniprot gffutils database; if not specified, '
                'a miniprot database will be built automatically'
    )
    parser_gffutils_grp = args_gffutils(parser)
    parser.add_argument('-ad', '--annotation-database', metavar='SOURCE', help='The source of the reference annotation (RefSeq / GENCODE / others).', default = "RefSeq")
    parser.add_argument('--no-auto-convert-gtf', action='store_true', default=False,
                        help='Disable automatic GTF to GFF3 conversion. By default, LiftOn will attempt to convert GTF files to GFF3 for better compatibility.')
    ###################################
    # END for the LiftOn params
    ###################################
    parser._positionals.title = '* Required input (sequences)'
    parser._optionals.title = '* Miscellaneous settings'
    parser._action_groups = [parser._positionals, referencegrp, referenceseqgrp, 
                             liftoffrefrgrp, miniprotrefrgrp, parser_gffutils_grp, 
                             parser_outgrp, parser._optionals, parser_aligngrp]
    args = parser.parse_args(arglist)
    if '-a' not in args.mm2_options:
        args.mm2_options += ' -a'
    if '--eqx' not in args.mm2_options:
        args.mm2_options += ' --eqx'
    if '-N' not in args.mm2_options:
        args.mm2_options += " -N 50"
    if '-p' not in args.mm2_options:
        args.mm2_options += " -p 0.5"
    if '--end-bonus' not in args.mm2_options:
        args.mm2_options += " --end-bonus 5"
    if (float(args.s) > float(args.sc)):
        parser.error("-sc must be greater than or equal to -s")
    if (args.chroms is None and args.unplaced is not None):
        parser.error("-unplaced must be used with -chroms")    
    return args
    

def run_all_lifton_steps(args):
    t1 = time.process_time()
    ################################
    # Step 0: Reading target & reference genomes
    ################################
    tgt_genome = args.target
    ref_genome = args.reference
    if args.output == "stdout": 
        outdir = "."
    else:
        outdir = os.path.dirname(args.output)
        outdir = outdir if outdir != "" else "."
        os.makedirs(outdir, exist_ok=True)
    lifton_outdir = f"{outdir}/lifton_output/"
    args.directory = "intermediate_files/"
    intermediate_dir = f"{outdir}/lifton_output/{args.directory}"
    os.makedirs(intermediate_dir, exist_ok=True)
    stats_dir= f"{outdir}/lifton_output/stats/"
    os.makedirs(stats_dir, exist_ok=True)
    args.directory = intermediate_dir
    logger.log(">> Reading target genome ...", debug=True)
    tgt_fai = Fasta(tgt_genome)
    logger.log(">> Reading reference genome ...", debug=True)
    ref_fai = Fasta(ref_genome)    

    t2 = time.process_time()
    ################################
    # Step 1: Building database from the reference annotation
    ################################
    logger.log("\n>> Creating reference annotation database : ", args.reference_annotation, debug=True)
    auto_convert_gtf = not args.no_auto_convert_gtf
    ref_db = annotation.Annotation(args.reference_annotation, args.infer_genes, args.infer_transcripts, args.merge_strategy, args.id_spec, args.force, args.verbose, auto_convert_gtf)

    t3 = time.process_time()
    ################################
    # Step 2: Get all reference features to liftover
    ################################
    features = lifton_utils.get_parent_features_to_lift(args.features)
    ref_features_dict, ref_features_len_dict, ref_features_reverse_dict, ref_trans_exon_num_dict = lifton_utils.get_ref_liffover_features(features, ref_db, intermediate_dir, args)

    t4 = time.process_time()
    ################################
    # Step 3: Extract protein & DNA dictionaries from the selected reference features
    ################################
    ref_trans_file = args.transcripts    
    ref_proteins_file = args.proteins    
    if (ref_proteins_file is None) or (not os.path.exists(ref_proteins_file)) or (ref_trans_file is None) or (not os.path.exists(ref_trans_file)):
        logger.log(">> Creating transcript DNA dictionary from the reference annotation ...", debug=True)
        logger.log(">> Creating transcript protein dictionary from the reference annotation ...", debug=True)
        ref_trans, ref_proteins = extract_sequence.extract_features(ref_db, features, ref_fai)
        ref_proteins_file = lifton_utils.write_seq_2_file(intermediate_dir, ref_proteins, "proteins")
        ref_trans_file = lifton_utils.write_seq_2_file(intermediate_dir, ref_trans, "transcripts")
    else:
        logger.log(">> Reading transcript DNA dictionary from the reference fasta ...", debug=True)
        logger.log(">> Reading transcript protein dictionary from the reference fasta ...", debug=True)
        ref_trans = Fasta(ref_trans_file)
        ref_proteins = Fasta(ref_proteins_file)
    logger.log("\t * number of transcripts: ", len(ref_trans.keys()), debug=True)
    logger.log("\t * number of proteins: ", len(ref_proteins.keys()), debug=True)
    trunc_ref_proteins = lifton_utils.get_truncated_protein(ref_proteins)
    logger.log("\t\t * number of truncated proteins: ", len(trunc_ref_proteins.keys()), debug=True)

    ################################
    # optional Step: Evaluation mode
    ################################
    if args.evaluation:
        tgt_annotation = args.output
        ref_annotation = args.reference_annotation
        print("Run LiftOn in evaluation mode")
        print("lifton_outdir     : ", lifton_outdir)
        print("Ref genome        : ", ref_genome)
        print("Target genome     : ", tgt_genome)
        print("Ref annotation    : ", args.reference_annotation)
        print("Target annotation : ", args.output)
        print("ref_trans_file    : ", ref_trans_file)
        print("ref_proteins_file : ", ref_proteins_file)
        logger.log(">> Creating target database : ", tgt_annotation, debug=True)
        os.makedirs(lifton_outdir, exist_ok=True)
        auto_convert_gtf = not args.no_auto_convert_gtf
        tgt_feature_db = annotation.Annotation(tgt_annotation, args.infer_genes, args.infer_transcripts, args.merge_strategy, args.id_spec, args.force, args.verbose, auto_convert_gtf).db_connection
        fw_score = open(lifton_outdir+"/eval.txt", "w")
        tree_dict = {}
        processed_features = 0
        for feature in features:
            for locus in tgt_feature_db.features_of_type(feature):#, limit=("chr1", 146652669, 146708545)):
                lifton_gene = run_evaluation.evaluation(None, locus, ref_db.db_connection, tgt_feature_db, tree_dict, tgt_fai, ref_proteins, ref_trans, ref_features_dict, fw_score, args, ENTRY_FEATURE=True)
                if processed_features % 20 == 0:
                    sys.stdout.write("\r>> LiftOn evaluated: %i features." % processed_features)
                processed_features += 1
        fw_score.close()
        return
    
    ################################
    # Step 4: Run liftoff & miniprot
    ################################
    t5 = time.process_time()
    liftoff_annotation = lifton_utils.exec_liftoff(lifton_outdir, ref_db, args)
    t6 = time.process_time()
    miniprot_annotation = lifton_utils.exec_miniprot(lifton_outdir, args, tgt_genome, ref_proteins_file)

    t7 = time.process_time()
    ################################
    # Step 5: Create liftoff and miniprot database
    ################################
    logger.log(f"\n>> Creating liftoff annotation database : {liftoff_annotation}", debug=True)
    auto_convert_gtf = not args.no_auto_convert_gtf
    l_feature_db = annotation.Annotation(liftoff_annotation, args.infer_genes, args.infer_transcripts, args.merge_strategy, args.id_spec, args.force, args.verbose, auto_convert_gtf).db_connection
    t8 = time.process_time()
    logger.log(f">> Creating miniprot annotation database : {miniprot_annotation}", debug=True)
    m_feature_db = annotation.Annotation(miniprot_annotation, args.infer_genes, args.infer_transcripts, args.merge_strategy, args.id_spec, args.force, args.verbose, auto_convert_gtf).db_connection
    fw = open(args.output, "w")
    fw_score = open(f"{lifton_outdir}/score.txt", "w")
    fw_unmapped = open(f"{stats_dir}/unmapped_features.txt", "w")
    fw_extra_copy = open(f"{stats_dir}/extra_copy_features.txt", "w")
    fw_mapped_feature = open(f'{stats_dir}/mapped_feature.txt', 'w')
    fw_mapped_trans = open(f'{stats_dir}/mapped_transcript.txt', 'w')
    fw_chain = open(f"{lifton_outdir}/chain.txt", "w") if args.write_chains else None

    t9 = time.process_time()
    ################################
    # Step 6: Creating miniprot 2 Liftoff ID mapping & Initializing intervaltree
    ################################
    ref_id_2_m_id_trans_dict, m_id_2_ref_id_trans_dict = lifton_utils.miniprot_id_mapping(m_feature_db)
    tree_dict = intervals.initialize_interval_tree(l_feature_db, features)
    transcripts_stats_dict = {'coding': {}, 'non-coding': {}, 'other': {}}
    processed_features = 0
    
    t10 = time.process_time()
    ################################
    # Step 7: Process Liftoff genes & transcripts
    #     structure 1: gene -> transcript -> exon
    #     structure 2: transcript -> exon
    ################################
    for feature in features:
        for locus in l_feature_db.features_of_type(feature):
            lifton_gene = run_liftoff.process_liftoff(None, locus, ref_db.db_connection, l_feature_db, ref_id_2_m_id_trans_dict, m_feature_db, tree_dict, tgt_fai, ref_proteins, ref_trans, ref_features_dict, fw_score, fw_chain, args, ENTRY_FEATURE=True)
            if lifton_gene is None or lifton_gene.ref_gene_id is None:
                continue
            lifton_gene.write_entry(fw,transcripts_stats_dict)
            if processed_features % 20 == 0:
                sys.stdout.write("\r>> LiftOn processed: %i features." % processed_features)
            processed_features += 1

    t11 = time.process_time()
    ################################
    # Step 8: Process miniprot transcripts
    ################################
    for mtrans in m_feature_db.features_of_type('mRNA'):
        lifton_gene = run_miniprot.process_miniprot(mtrans, ref_db, m_feature_db, tree_dict, tgt_fai, ref_proteins, ref_trans, ref_features_dict, fw_score, m_id_2_ref_id_trans_dict, ref_features_len_dict, ref_trans_exon_num_dict, ref_features_reverse_dict, args)
        if lifton_gene is None or lifton_gene.ref_gene_id is None:
            continue
        lifton_gene.write_entry(fw, transcripts_stats_dict)
        if processed_features % 20 == 0:
            sys.stdout.write("\r>> LiftOn processed: %i features." % processed_features)
        processed_features += 1
    
    t12 = time.process_time()
    ################################
    # Step 9: Printing stats
    ################################
    stats.print_report(ref_features_dict, transcripts_stats_dict, fw_unmapped, fw_extra_copy, fw_mapped_feature, fw_mapped_trans, debug=args.debug)
    fw.close()
    fw_score.close()
    fw_unmapped.close()
    fw_extra_copy.close()
    fw_mapped_feature.close()
    fw_mapped_trans.close()
    if args.write_chains: fw_chain.close()
    t13 = time.process_time()

    if args.measure_time:
        reading_target_reference_genomes= t2 - t1
        creating_reference_annotation_database = t3 - t2
        get_all_reference_features_to_lift = t4 - t3
        extract_protein_dna_dictionaries = t5 - t4
        run_liftoff_miniprot = t6 - t5
        create_liftoff_database = t8 - t6
        create_miniprot_database = t9 - t8
        miniprot_2_liftoff_id_mapping = t10 - t9
        process_liftoff_genes_transcripts = t11 - t10
        process_miniprot_transcripts = t12 - t11
        report_stats = t13 - t12
        overall_time = t13 - t1
        print("Time taken for each step:")
        print(f"Reading target & reference genomes: {reading_target_reference_genomes}")
        print(f"Creating reference annotation database: {creating_reference_annotation_database}")
        print(f"Get all reference features to liftover: {get_all_reference_features_to_lift}")
        print(f"Extract protein & DNA dictionaries: {extract_protein_dna_dictionaries}")
        print(f"Run liftoff & miniprot: {run_liftoff_miniprot}")
        print(f"Create liftoff database: {create_liftoff_database}")
        print(f"Create miniprot database: {create_miniprot_database}")
        print(f"Miniprot 2 Liftoff ID mapping: {miniprot_2_liftoff_id_mapping}")
        print(f"Process Liftoff genes & transcripts: {process_liftoff_genes_transcripts}")
        print(f"Process miniprot transcripts: {process_miniprot_transcripts}")
        print(f"Report stats: {report_stats}")
        print(f"Overall time: {overall_time}")
        fw_time = open(f"{outdir}/time.txt", "w")
        fw_time.write(f"{reading_target_reference_genomes}\tReading target & reference genomes\n")
        fw_time.write(f"{creating_reference_annotation_database}\tCreating reference annotation database\n")
        fw_time.write(f"{get_all_reference_features_to_lift}\tGet all reference features to liftover\n")
        fw_time.write(f"{extract_protein_dna_dictionaries}\tExtract protein & DNA dictionaries\n")
        fw_time.write(f"{run_liftoff_miniprot}\tRun liftoff & miniprot\n")
        fw_time.write(f"{create_liftoff_database}\tCreate liftoff database\n")
        fw_time.write(f"{create_miniprot_database}\tCreate miniprot database\n")
        fw_time.write(f"{miniprot_2_liftoff_id_mapping}\tMiniprot 2 Liftoff ID mapping\n")
        fw_time.write(f"{process_liftoff_genes_transcripts}\tProcess Liftoff genes & transcripts\n")
        fw_time.write(f"{process_miniprot_transcripts}\tProcess miniprot transcripts\n")
        fw_time.write(f"{report_stats}\tReport stats\n")
        fw_time.write(f"{overall_time}\tOverall time\n")
        fw_time.close()


def main(arglist=None):
    banner = '''
====================================================================
An accurate homology lift-over tool between assemblies
====================================================================


    ██╗     ██╗███████╗████████╗ ██████╗ ███╗   ██╗
    ██║     ██║██╔════╝╚══██╔══╝██╔═══██╗████╗  ██║
    ██║     ██║█████╗     ██║   ██║   ██║██╔██╗ ██║
    ██║     ██║██╔══╝     ██║   ██║   ██║██║╚██╗██║
    ███████╗██║██║        ██║   ╚██████╔╝██║ ╚████║
    ╚══════╝╚═╝╚═╝        ╚═╝    ╚═════╝ ╚═╝  ╚═══╝
    '''
    print(banner, file=sys.stderr)
    args = parse_args(arglist)
    if not run_miniprot.check_miniprot_installed(): 
        sys.exit("miniprot is not installed. Please install miniprot before running LiftOn.")
    run_all_lifton_steps(args)



def has_stop_codon(ref_align, target_align):
    """
        This function checks if a string contains a stop codon.

        Parameters:
        - ref_align: reference sequence alignment
        - target_align: target sequence alignment

        Returns:
        True if a stop codon is found, False otherwise.
    """
    for i, letter in enumerate(target_align):
        if letter == "*" and ref_align[i] != "*":
            return True
    return False


def is_frameshift(s):
    """
        This function checks if a string contains a non-divisible substring of '-' characters.

        Parameters:
        - s: a string

        Returns:
        True if s contains a non-divisible substring of '-' characters, False otherwise.
    """
    # Initialize a variable to keep track of consecutive '-' characters.
    consecutive_count = 0
    for char in s:
        if char == '-':
            consecutive_count += 1
        else:
            # If we encounter a non-'-' character, check if the consecutive count is not divisible by three.
            if consecutive_count % 3 != 0:
                return True
            consecutive_count = 0
    # After the loop, check the last substring if it's not divisible by three.
    if consecutive_count % 3 != 0:
        return True    
    # If no non-divisible substrings are found, return False.
    return False


def find_variants(align_dna, align_protein, lifton_status, peps, is_non_coding):
    """
        This function finds the variants between two sequences.

        Parameters:
        - align_dna: DNA pairwise alignment
        - align_protein: Protein pairwise alignment
        - lifton_status: Lifton_Status object
        - peps: Protein sequence split by stop codon

        Returns:
        None
    """
    # Mutation types:
    #   (1) identical
    #   (2) synonymous
    #   (3) frameshift
    #   (4) start_lost
    #   (5) inframe_insertion
    #   (6) inframe_deletion
    #   (7) nonsynonymous
    #   (8) stop_missing
    #   (9) stop_codon_gain
    mutation_type = []
    if is_non_coding:
        mutation_type.append('non_coding')
        lifton_status.status = mutation_type
        return
    if align_dna == None:
        mutation_type.append('full_transcript_loss')
        lifton_status.status = mutation_type
        return 
    if align_protein == None:
        mutation_type.append('no_protein')
        lifton_status.status = mutation_type
        return 
    # 1. return cases
    if align_dna.identity == 1.0:
        mutation_type.append('identical')
        lifton_status.status = mutation_type
        return 
    # 2. return cases
    frameshift = False
    if align_protein.identity == 1.0:
        mutation_type.append('synonymous')
        lifton_status.status = mutation_type
        return
    # 3. return cases
    if is_frameshift(align_dna.query_aln):
        mutation_type.append('frameshift')
        frameshift = True
    if is_frameshift(align_dna.ref_aln) and frameshift == False:
        mutation_type.append('frameshift')
        frameshift = True
    if align_dna.query_aln[0:3] != align_dna.ref_aln[0:3] and \
        align_dna.query_aln[0:3] != 'ATG' and \
            align_protein.query_aln[0] != align_protein.ref_aln[0] and \
                align_protein.query_aln[0] != "M" :
        mutation_type.append("start_lost")
    if "-" in align_dna.ref_aln and not frameshift:
        mutation_type.append("inframe_insertion")
    if "-" in align_dna.query_aln and not frameshift:
        mutation_type.append("inframe_deletion")
    if len(peps) == 2 and str(peps[1]) == "":
        # This is a valid protein ends with stop codon *
        # But alignment is not 100% identical
        if len(mutation_type) == 0:
            mutation_type.append("nonsynonymous")
    elif len(peps) == 1:
        # This is a protein without stop codon
        mutation_type.append("stop_missing")
    else:
        mutation_type.append("stop_codon_gain")
        stop_codon_count = 0
        for idx, ele in enumerate(align_protein.query_seq):
            if ele == "*" and idx != len(align_protein.query_seq)-1:
                stop_codon_count += 1
    lifton_status.status = mutation_type


import sys 

def log(*argv, debug=False):
    """
        This function logs the message to stderr.

        Parameters:
        - argv: message
        - debug: debug mode
    """
    if debug:
        print(*argv, file=sys.stderr)



from Bio.Seq import Seq

def determine_file_format(file_path):
    """Detect if the input file is GTF or GFF3 format.
    
    GTF format uses space-separated key-value pairs with gene_id and transcript_id.
    GFF3 format uses semicolon-separated key=value pairs with ID and Parent.
    
    Default format is GFF3 (most common and recommended format).
    
    Returns:
        "GTF format" if GTF format is detected
        "GFF format" if GFF3 format is detected (default)
        "Unknown format" if format cannot be determined (should be treated as GFF3)
    """
    gff_keys = {"ID", "Parent", "Name"}
    gtf_keys = {"gene_id", "transcript_id"}
    gtf_count = 0
    gff_count = 0
    lines_checked = 0
    max_lines_to_check = 100  # Check first 100 non-comment lines
    
    try:
        with open(file_path, 'r') as file:
            for line in file:
                # Skip header or comment lines
                if line.startswith("#"):
                    continue
                
                columns = line.strip().split("\t")
                if len(columns) != 9:
                    # Not a valid GTF/GFF file line, skip
                    continue
                
                attributes_column = columns[8]
                
                # GTF format: key-value pairs separated by spaces and semicolons
                # e.g., "gene_id \"ENSG00000139618\"; transcript_id \"ENST00000380152\";"
                # Check for GTF-specific patterns (key followed by space and quoted value)
                # GTF uses: gene_id "value"; transcript_id "value";
                if any((key + " " in attributes_column or key + "\"" in attributes_column) and 
                       not (key + "=" in attributes_column) for key in gtf_keys):
                    gtf_count += 1
                
                # GFF3 format: key=value pairs separated by semicolons
                # e.g., "ID=gene1;Parent=;Name=Gene1"
                # Check for GFF3-specific patterns (key=value)
                if any(key + "=" in attributes_column for key in gff_keys):
                    gff_count += 1
                
                lines_checked += 1
                if lines_checked >= max_lines_to_check:
                    break
        
        # Determine format based on counts
        # GTF format is more specific, so if we see GTF patterns (without =), it's likely GTF
        # Default to GFF3 format (most common and recommended)
        if gtf_count > 0 and gtf_count > gff_count:
            return "GTF format"
        elif gff_count > 0:
            return "GFF format"
        else:
            # No clear indicators found, default to GFF3 (most common format)
            return "GFF format"
            
    except Exception as e:
        # If file cannot be read, default to GFF3 format
        return "GFF format"


def extract_features(ref_db, features, ref_fai):
    ref_trans = {}
    ref_proteins = {}
    counter = 0
    for feature in features:
        for locus in ref_db.db_connection.features_of_type(feature):
            # print(f"Extracting features for {locus.id}")
            counter += 1
            __inner_extract_feature(ref_db, locus, ref_fai, ref_trans, ref_proteins)
    print(f"Extracted features for {counter} features")
    return ref_trans, ref_proteins


def __inner_extract_feature(ref_db, feature, ref_fai, ref_trans, ref_proteins):
    # If exon is the first level children
    children_exons = list(ref_db.db_connection.children(feature, featuretype='exon', level=1))
    children_CDSs = list(ref_db.db_connection.children(feature, featuretype=('start_codon', 'CDS', 'stop_codon'), level=1))
    # print(f"Parent: {feature.id};  exon: {len(children_exons)}; CDS: {len(children_CDSs)}")
    if len(children_exons) > 0 or len(children_CDSs) > 0:
        # print(f"Parent: {feature.id};  exon: {len(children_exons)}; CDS: {len(children_CDSs)}")
        if len(children_exons) > 0:
            trans_seq = get_dna_sequence(feature, ref_fai, children_exons)
            trans_seq = trans_seq.upper()
            ref_trans[feature.id] = trans_seq
        if len(children_CDSs) > 0:
            protein_seq = get_protein_sequence(feature, ref_fai, children_CDSs)
            protein_seq = protein_seq.upper()
            ref_proteins[feature.id] = protein_seq
    else:
        for child in ref_db.db_connection.children(feature, level=1, order_by='start'):
            __inner_extract_feature(ref_db, child, ref_fai, ref_trans, ref_proteins)
            

def merge_children_intervals(children):
    if not children:
        return []
    intervals = [[child.start, child.end] for child in children]
    intervals.sort(key=lambda interval: interval[0])
    merged = [intervals[0]]
    for current in intervals[1:]:
        previous = merged[-1]
        if current[0] <= previous[1]:
            previous[1] = max(previous[1], current[1])
        else:
            merged.append(current)
    return merged


def get_dna_sequence(parent_feature, fasta, features):
    chrom = parent_feature.seqid
    strand = parent_feature.strand
    merged_features = merge_children_intervals(features)
    sequence =''
    if chrom not in fasta.keys():
        return sequence
    for start, end in merged_features:
        sequence += str(fasta[chrom][start -1: end])
    if strand == "-":
        sequence = str(Seq(sequence).reverse_complement())
    sequence += 'N' * get_padding_length(len(sequence))
    return sequence


def get_padding_length(sequence_length):
    return (3-sequence_length%3)%3


def get_protein_sequence(parent_feature, fasta, features):
    dna = get_dna_sequence(parent_feature, fasta, features)
    return str(Seq(dna).translate())


from lifton.liftoff  import find_best_mapping, liftoff_utils, merge_lifted_features


def lift_all_features(alns, threshold, feature_db,  feature_hierarchy,
                      unmapped_features, lifted_feature_list, seq_id_threshold, feature_locations, args,
                      ref_parent_order):
    features_to_lift = feature_hierarchy.parents
    feature_order = get_feature_order(feature_db)
    alignments = sort_alignments(features_to_lift, alns)
    num_features = 0
    for alignment in alignments:
        num_features += 1
        preivous_feature_start, previous_feature_seq, previous_feature_ref_start = find_neighbor_location(
            feature_hierarchy.parents, alignment, lifted_feature_list, ref_parent_order)

        lifted_features, parent_name = lift_single_feature(threshold, feature_order, features_to_lift,
                                                           feature_hierarchy, preivous_feature_start,
                                                           previous_feature_ref_start,
                                                           previous_feature_seq, unmapped_features,
                                                           alignment, seq_id_threshold, feature_locations,
                                                           lifted_feature_list, args)

        if lifted_features != []:
            lifted_feature_list[parent_name] = lifted_features



def get_feature_order(gene_db):
    feature_types = list(gene_db.featuretypes())
    index = 0
    feature_order = {}
    if 'exon' in feature_types:
        feature_order['exon'] = index
        index += 1
    if 'CDS' in feature_types:
        feature_order['CDS'] = index
        index += 1
    for feature_type in feature_types:
        if feature_type not in feature_order:
            feature_order[feature_type] = index
            index += 1
    return feature_order


def sort_alignments(parent_dict, alignments):
    parent_list = []
    order = 0
    order_dict = {}
    values = list(alignments.values())
    for alignment in alignments:
        parent_list.append(parent_dict[liftoff_utils.convert_id_to_original(alignments[alignment][0].query_name)])
    parent_list.sort(key=lambda x: (x.seqid, x.start))
    for parent in parent_list:
        order_dict[parent.id] = order
        order += 1
    values.sort(key=lambda x: order_dict[liftoff_utils.convert_id_to_original(x[0].query_name)])
    return values


def find_neighbor_location(ref_parents, alignment, lifted_feature_list, ref_parent_order):
    ref_feature = ref_parents[liftoff_utils.convert_id_to_original(alignment[0].query_name)]
    ref_neighbor_name = liftoff_utils.find_nonoverlapping_upstream_neighbor(ref_parent_order, ref_feature.id)
    if ref_neighbor_name is not None:
        ref_neighbor_key = ref_neighbor_name + "_0"
        if ref_neighbor_key in lifted_feature_list:
            previous_feature_start = lifted_feature_list[ref_neighbor_key][0].start
            previous_feature_seq = lifted_feature_list[ref_neighbor_key][0].seqid
            previous_feature_ref_start = ref_parents[ref_neighbor_name].start
            return previous_feature_start, previous_feature_seq, previous_feature_ref_start
    return 0, "", 0


def lift_single_feature(threshold, feature_order, features_to_lift, feature_hierarchy,
                        previous_feature_start, previous_feature_ref_start, previous_gene_seq, unmapped_features,
                        aligned_feature, seq_id_threshold,
                        feature_locations,
                        lifted_features_list, args):
    new_parent_name = aligned_feature[0].query_name
    original_parent_name = liftoff_utils.convert_id_to_original(new_parent_name)
    parent = features_to_lift[original_parent_name]
    if len(aligned_feature) > 0:
        lifted_children, alignment_coverage, seq_id = find_best_mapping.find_best_mapping(aligned_feature,
                                                                                          parent.end - parent.start + 1,
                                                                                          parent,
                                                                                          feature_hierarchy,
                                                                                          previous_feature_start,
                                                                                          previous_feature_ref_start,
                                                                                          previous_gene_seq,
                                                                                          feature_locations,
                                                                                          lifted_features_list,
                                                                                          args)


        lifted_features = merge_lifted_features.merge_lifted_features(lifted_children,
                                                                      parent,
                                                                      unmapped_features, threshold,
                                                                      new_parent_name, feature_order,
                                                                      feature_hierarchy,
                                                                      alignment_coverage, seq_id,
                                                                      seq_id_threshold)



    else:
        unmapped_features.append(parent)
    return lifted_features, aligned_feature[0].query_name



from intervaltree import Interval, IntervalTree

def initialize_interval_tree(l_feature_db, features):
    tree_dict = {}
    # Adding gene intervals into intervaltree
    for feature in features:
        for locus in l_feature_db.features_of_type(feature):#, limit=("chrX", 0, 200914237)):
            gene_interval = Interval(locus.start, locus.end, locus.id)
            chromosome = locus.seqid
            if chromosome not in tree_dict.keys():
                tree_dict[chromosome] = IntervalTree()
            tree_dict[chromosome].add(gene_interval)
    return tree_dict